# SPDX-License-Identifier: BSD-2-Clause

from .gray     import Encoder as GrayEncoder, Decoder as GrayDecoder
from .one_hot  import Encoder, Decoder
from .priority import Encoder as PriorityEncoder, Decoder as PriorityDecoder

__all__ = (
	'Encoder',
	'Decoder',
	'GrayEncoder',
	'GrayDecoder',
	'PriorityEncoder',
	'PriorityDecoder',
)



# SPDX-License-Identifier: BSD-2-Clause


from torii.lib.soc.periph import *
from torii.lib.soc.memory import MemoryMap
from torii.lib.soc        import event

from ...utils             import ToriiTestSuiteCase

class ConstantBoolTestCase(ToriiTestSuiteCase):
	def test_init(self):
		a = ConstantBool(True)
		b = ConstantBool(False)
		self.assertTrue(a.value)
		self.assertFalse(b.value)

	def test_value_wrong(self):
		with self.assertRaisesRegex(TypeError, r'Value must be a bool, not \'foo\''):
			ConstantBool('foo')

	def test_repr(self):
		self.assertEqual(repr(ConstantBool(True)), 'ConstantBool(True)')


class ConstantIntTestCase(ToriiTestSuiteCase):
	def test_init(self):
		c = ConstantInt(5, width = 8, signed = True)
		self.assertEqual(c.value,  5)
		self.assertEqual(c.width,  8)
		self.assertEqual(c.signed, True)

	def test_init_default(self):
		c = ConstantInt(5)
		self.assertEqual(c.value,  5)
		self.assertEqual(c.width,  3)
		self.assertEqual(c.signed, False)

	def test_value_wrong(self):
		with self.assertRaisesRegex(TypeError, r'Value must be an integer, not \'foo\''):
			ConstantInt('foo')

	def test_width_wrong(self):
		with self.assertRaisesRegex(TypeError, r'Width must be an integer, not \'foo\''):
			ConstantInt(5, width = "foo")

	def test_width_overflow(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Width must be greater than or equal to the number of bits needed to represent 5'
		):
			ConstantInt(5, width = 1)

	def test_signed_wrong(self):
		with self.assertRaisesRegex(TypeError, r'Signedness must be a bool, not \'foo\''):
			ConstantInt(5, signed = "foo")

	def test_repr(self):
		self.assertEqual(
			repr(ConstantInt(-5, width = 8, signed = True)),
			'ConstantInt(-5, width=8, signed=True)'
		)


class ConstantMapTestCase(ToriiTestSuiteCase):
	def test_init(self):
		constant_map = ConstantMap(A = 5, B = True, C = ConstantBool(False))
		self.assertEqual(
			repr(constant_map), 'ConstantMap(['
			'(\'A\', ConstantInt(5, width=3, signed=False)), '
			'(\'B\', ConstantBool(True)), '
			'(\'C\', ConstantBool(False))])',
		)

	def test_init_wrong_value(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Constant value must be an instance of ConstantValue, not \(\'foo\', \'bar\'\)'
		):
			ConstantMap(A = ('foo', 'bar'))

	def test_getitem(self):
		a = ConstantInt(1)
		b = ConstantBool(False)
		constant_map = ConstantMap(A = a, B = b)
		self.assertIs(constant_map['A'], a)
		self.assertIs(constant_map['B'], b)

	def test_iter(self):
		a = ConstantInt(1)
		b = ConstantBool(False)
		constant_map = ConstantMap(B = b, A = a)
		self.assertEqual(list(constant_map.items()), [
			('B', b),
			('A', a),
		])

	def test_len(self):
		a = ConstantInt(1)
		b = ConstantBool(False)
		constant_map = ConstantMap(B = b, A = a)
		self.assertEqual(len(constant_map), 2)


class PeripheralInfoTestCase(ToriiTestSuiteCase):
	def test_memory_map(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		info = PeripheralInfo(memory_map = memory_map)
		self.assertIs(info.memory_map, memory_map)

	def test_memory_map_frozen(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		PeripheralInfo(memory_map = memory_map)
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has been frozen. Cannot add resource \'a\''
		):
			memory_map.add_resource('a', name = 'foo', size = 3, extend = True)

	def test_memory_map_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Memory map must be an instance of MemoryMap, not \'foo\''
		):
			PeripheralInfo(memory_map = 'foo')

	def test_irq(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		irq = event.Source()
		info = PeripheralInfo(memory_map = memory_map, irq = irq)
		self.assertIs(info.irq, irq)

	def test_irq_none(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		info = PeripheralInfo(memory_map = memory_map, irq = None)
		with self.assertRaisesRegex(
			NotImplementedError,
			r'Peripheral info does not have an IRQ line'
		):
			info.irq

	def test_irq_default(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		info = PeripheralInfo(memory_map = memory_map)
		with self.assertRaisesRegex(
			NotImplementedError,
			r'Peripheral info does not have an IRQ line'
		):
			info.irq

	def test_irq_wrong(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'IRQ line must be an instance of event.Source, not \'foo\''
		):
			PeripheralInfo(memory_map = memory_map, irq = 'foo')

	def test_constant_map(self):
		constant_map = ConstantMap()
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		info = PeripheralInfo(memory_map = memory_map, constant_map = constant_map)
		self.assertIs(info.constant_map, constant_map)

	def test_constant_map_none(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		info = PeripheralInfo(memory_map = memory_map, constant_map = None)
		self.assertIsInstance(info.constant_map, ConstantMap)
		self.assertEqual(info.constant_map, {})

	def test_constant_map_default(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		info = PeripheralInfo(memory_map = memory_map)
		self.assertIsInstance(info.constant_map, ConstantMap)
		self.assertEqual(info.constant_map, {})

	def test_constant_map_wrong(self):
		memory_map = MemoryMap(addr_width = 1, data_width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'Constant map must be an instance of ConstantMap, not \'foo\''
		):
			PeripheralInfo(memory_map = memory_map, constant_map = 'foo')



from torii import *


class UpCounter(Elaboratable):
    '''
    A 16-bit up counter with a fixed limit.

    Parameters
    ----------
    limit : int
        The value at which the counter overflows.

    Attributes
    ----------
    en : Signal, in
        The counter is incremented if ``en`` is asserted, and retains
        its value otherwise.
    ovf : Signal, out
        ``ovf`` is asserted when the counter reaches its limit.
    '''
    def __init__(self, limit: int):
        self.limit = limit

        # Ports
        self.en  = Signal()
        self.ovf = Signal()

        # State
        self.count = Signal(16)

    def elaborate(self, platform) -> Module:
        m = Module()

        m.d.comb += self.ovf.eq(self.count == self.limit)

        with m.If(self.en):
            with m.If(self.ovf):
                m.d.sync += self.count.eq(0)
            with m.Else():
                m.d.sync += self.count.eq(self.count + 1)

        return m
# --- TEST ---
from torii.sim import Simulator


dut = UpCounter(25)
def bench():
    # Disabled counter should not overflow.
    yield dut.en.eq(0)
    for _ in range(30):
        yield
        assert not (yield dut.ovf)

    # Once enabled, the counter should overflow in 25 cycles.
    yield dut.en.eq(1)
    for _ in range(25):
        yield
        assert not (yield dut.ovf)
    yield
    assert (yield dut.ovf)

    # The overflow should clear in one cycle.
    yield
    assert not (yield dut.ovf)


sim = Simulator(dut)
sim.add_clock(1e-6) # 1 MHz
sim.add_sync_process(bench)
with sim.write_vcd('up_counter.vcd'):
    sim.run()
# --- CONVERT ---
from torii.back import verilog


top = UpCounter(25)
with open('up_counter.v', 'w') as f:
    f.write(verilog.convert(top, ports = [top.en, top.ovf]))



# SPDX-License-Identifier: BSD-2-Clause

from abc        import abstractmethod
from typing     import Literal, Optional

from ....build  import Attrs, Clock, Subsignal, TemplatedPlatform
from ....hdl    import (
	ClockDomain, ClockSignal, Const, Instance, Module, Record, Signal
)
from ....lib.io import Pin

__all__ = (
	'ECP5Platform',
)

class ECP5Platform(TemplatedPlatform):
	'''
	.. rubric:: Trellis toolchain

	Required tools:
		* ``yosys``
		* ``nextpnr-ecp5``
		* ``ecppack``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_TRELLIS``, if present.

	Available overrides:
		* ``verbose``: enables logging of informational messages to standard error.
		* ``read_verilog_opts``: adds options for ``read_verilog`` Yosys command.
		* ``synth_opts``: adds options for ``synth_ecp5`` Yosys command.
		* ``script_after_read``: inserts commands after ``read_rtlil`` in Yosys script.
		* ``script_after_synth``: inserts commands after ``synth_ecp5`` in Yosys script.
		* ``yosys_opts``: adds extra options for ``yosys``.
		* ``nextpnr_opts``: adds extra options for ``nextpnr-ecp5``.
		* ``ecppack_opts``: adds extra options for ``ecppack``.
		* ``add_preferences``: inserts commands at the end of the LPF file.

	Build products:
		* ``{{name}}.rpt``: Yosys log.
		* ``{{name}}.json``: synthesized RTL.
		* ``{{name}}.tim``: nextpnr log.
		* ``{{name}}.config``: ASCII bitstream.
		* ``{{name}}.bit``: binary bitstream.
		* ``{{name}}.svf``: JTAG programming vector.

	.. rubric:: Diamond toolchain

	Required tools:
		* ``pnmainc``
		* ``ddtcmd``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_DIAMOND``, if present. On Linux, diamond_env as provided by Diamond
	itself is a good candidate. On Windows, the following script (named ``diamond_env.bat``,
	for instance) is known to work::

		@echo off
		set PATH = C:\\lscc\\diamond\\%DIAMOND_VERSION%\\bin\\nt64;%PATH%

	Available overrides:
		* ``script_project``: inserts commands before ``prj_project save`` in Tcl script.
		* ``script_after_export``: inserts commands after ``prj_run Export`` in Tcl script.
		* ``add_preferences``: inserts commands at the end of the LPF file.
		* ``add_constraints``: inserts commands at the end of the XDC file.

	Build products:
		* ``{{name}}_impl/{{name}}_impl.htm``: consolidated log.
		* ``{{name}}.bit``: binary bitstream.
		* ``{{name}}.svf``: JTAG programming vector.

	'''

	toolchain = None # selected when creating platform

	@property
	@abstractmethod
	def device(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def package(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def speed(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	grade   = 'C' # [C]ommercial, [I]ndustrial

	# Trellis templates

	_nextpnr_device_options = {
		'LFE5U-12F':    '--12k',
		'LFE5U-25F':    '--25k',
		'LFE5U-45F':    '--45k',
		'LFE5U-85F':    '--85k',
		'LFE5UM-25F':   '--um-25k',
		'LFE5UM-45F':   '--um-45k',
		'LFE5UM-85F':   '--um-85k',
		'LFE5UM5G-25F': '--um5g-25k',
		'LFE5UM5G-45F': '--um5g-45k',
		'LFE5UM5G-85F': '--um5g-85k',
	}
	_nextpnr_package_options = {
		'BG256': 'caBGA256',
		'MG285': 'csfBGA285',
		'BG381': 'caBGA381',
		'BG554': 'caBGA554',
		'BG756': 'caBGA756',
	}

	_trellis_required_tools = [
		'yosys',
		'nextpnr-ecp5',
		'ecppack'
	]
	_trellis_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'{{name}}.il': r'''
			# {{autogenerated}}
			{{emit_rtlil()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.ys': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".v") -%}
				read_verilog {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".sv") -%}
				read_verilog -sv {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".il") -%}
				read_rtlil {{file}}
			{% endfor %}
			read_rtlil {{name}}.il
			delete w:$verilog_initial_trigger
			{{get_override("script_after_read")|default("# (script_after_read placeholder)")}}
			synth_ecp5 {{get_override("synth_opts")|options}} -top {{name}}
			{{get_override("script_after_synth")|default("# (script_after_synth placeholder)")}}
			write_json {{name}}.json
		''',
		'{{name}}.lpf': r'''
			# {{autogenerated}}
			BLOCK ASYNCPATHS;
			BLOCK RESETPATHS;
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				LOCATE COMP "{{port_name}}" SITE "{{pin_name}}";
				{% if attrs -%}
				IOBUF PORT "{{port_name}}"
					{%- for key, value in attrs.items() %} {{key}}={{value}}{% endfor %};
				{% endif %}
			{% endfor %}
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is not none -%}
					FREQUENCY PORT "{{port_signal.name}}" {{frequency}} HZ;
				{% else -%}
					FREQUENCY NET "{{net_signal|hierarchy(".")}}" {{frequency}} HZ;
				{% endif %}
			{% endfor %}
			{{get_override("add_preferences")|default("# (add_preferences placeholder)")}}
		'''
	}
	_trellis_command_templates = [
		r'''
		{{invoke_tool("yosys")}}
			{{quiet("-q")}}
			{{get_override("yosys_opts")|options}}
			-l {{name}}.rpt
			{{name}}.ys
		''',
		r'''
		{{invoke_tool("nextpnr-ecp5")}}
			{{quiet("--quiet")}}
			{{get_override("nextpnr_opts")|options}}
			--log {{name}}.tim
			{{platform._nextpnr_device_options[platform.device]}}
			--package {{platform._nextpnr_package_options[platform.package]|upper}}
			--speed {{platform.speed}}
			--json {{name}}.json
			--lpf {{name}}.lpf
			--textcfg {{name}}.config
		''',
		r'''
		{{invoke_tool("ecppack")}}
			{{verbose("--verbose")}}
			{{get_override("ecppack_opts")|options}}
			--input {{name}}.config
			--bit {{name}}.bit
			--svf {{name}}.svf
		'''
	]

	# Diamond templates

	_diamond_required_tools = [
		'pnmainc',
		'ddtcmd',
	]
	_diamond_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'build_{{name}}.sh': r'''
			# {{autogenerated}}
			set -e{{verbose("x")}}
			if [ -z "$BASH" ] ; then exec /bin/bash "$0" "$@"; fi
			if [ -n "${{platform._toolchain_env_var}}" ]; then
				bindir=$(dirname "${{platform._toolchain_env_var}}")
				. "${{platform._toolchain_env_var}}"
			fi
			{{emit_commands("sh")}}
		''',
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.tcl': r'''
			prj_project new -name {{name}} -impl impl -impl_dir {{name}}_impl \
				-dev {{platform.device}}-{{platform.speed}}{{platform.package}}{{platform.grade}} \
				-lpf {{name}}.lpf \
				-synthesis synplify
			{% for file in platform.iter_files(".v", ".sv", ".vhd", ".vhdl") -%}
				prj_src add {{file|tcl_escape}}
			{% endfor %}
			prj_src add {{name}}.v
			prj_impl option top {{name}}
			prj_src add {{name}}.sdc
			{{get_override("script_project")|default("# (script_project placeholder)")}}
			prj_project save
			prj_run Synthesis -impl impl
			prj_run Translate -impl impl
			prj_run Map -impl impl
			prj_run PAR -impl impl
			prj_run Export -impl impl -task Bitgen
			{{get_override("script_after_export")|default("# (script_after_export placeholder)")}}
		''',
		'{{name}}.lpf': r'''
			# {{autogenerated}}
			BLOCK ASYNCPATHS;
			BLOCK RESETPATHS;
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				LOCATE COMP "{{port_name}}" SITE "{{pin_name}}";
				{% if attrs -%}
				IOBUF PORT "{{port_name}}"
					{%- for key, value in attrs.items() %} {{key}}={{value}}{% endfor %};
				{% endif %}
			{% endfor %}
			{{get_override("add_preferences")|default("# (add_preferences placeholder)")}}
		''',
		'{{name}}.sdc': r'''
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is not none -%}
					create_clock -name {{port_signal.name|tcl_escape}} -period {{1000000000/frequency}} [get_ports {{port_signal.name|tcl_escape}}]
				{% else -%}
					create_clock -name {{net_signal.name|tcl_escape}} -period {{1000000000/frequency}} [get_nets {{net_signal|hierarchy("/")|tcl_escape}}]
				{% endif %}
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''' # noqa: E501
	}
	_diamond_command_templates = [
		# These don't have any usable command-line option overrides.
		r'''
		{{invoke_tool("pnmainc")}}
			{{name}}.tcl
		''',
		r'''
		{{invoke_tool("ddtcmd")}}
			-oft -bit
			-if {{name}}_impl/{{name}}_impl.bit -of {{name}}.bit
		''',
		r'''
		{{invoke_tool("ddtcmd")}}
			-oft -svfsingle -revd -op "Fast Program"
			-if {{name}}_impl/{{name}}_impl.bit -of {{name}}.svf
		''',
	]

	# Common logic

	def __init__(
		self, *, toolchain: Literal['Trellis', 'Diamond'] = 'Trellis'
	) -> None:
		super().__init__()

		assert toolchain in ('Trellis', 'Diamond')
		self.toolchain = toolchain

	@property
	def required_tools(self) -> list[str]:
		if self.toolchain == 'Trellis':
			return self._trellis_required_tools
		if self.toolchain == 'Diamond':
			return self._diamond_required_tools
		assert False

	@property
	def file_templates(self) -> dict[str, str]:
		if self.toolchain == 'Trellis':
			return self._trellis_file_templates
		if self.toolchain == 'Diamond':
			return self._diamond_file_templates
		assert False

	@property
	def command_templates(self) -> list[str]:
		if self.toolchain == 'Trellis':
			return self._trellis_command_templates
		if self.toolchain == 'Diamond':
			return self._diamond_command_templates
		assert False

	@property
	def default_clk_constraint(self) -> Clock:
		if self.default_clk == 'OSCG':
			return Clock(310e6 / self.oscg_div)
		return super().default_clk_constraint

	def create_missing_domain(self, name: str) -> Module:
		# Lattice ECP5 devices have two global set/reset signals: PUR, which is driven at startup
		# by the configuration logic and unconditionally resets every storage element, and GSR,
		# which is driven by user logic and each storage element may be configured as affected or
		# unaffected by GSR. PUR is purely asynchronous, so even though it is a low-skew global
		# network, its de-assertion may violate a setup/hold constraint with relation to a user
		# clock. To avoid this, a GSR/SGSR instance should be driven synchronized to user clock.
		if name == 'sync' and self.default_clk is not None:
			m = Module()
			if self.default_clk == 'OSCG':
				if not hasattr(self, 'oscg_div'):
					raise ValueError('OSCG divider (oscg_div) must be an integer between 2 and 128')
				if not isinstance(self.oscg_div, int) or self.oscg_div < 2 or self.oscg_div > 128:
					raise ValueError(
						f'OSCG divider (oscg_div) must be an integer between 2 and 128, not {self.oscg_div!r}'
					)
				clk_i = Signal()
				m.submodules += Instance('OSCG', p_DIV = self.oscg_div, o_OSC = clk_i)
			else:
				clk_i = self.request(self.default_clk).i
			if self.default_rst is not None:
				rst_i = self.request(self.default_rst).i
			else:
				rst_i = Const(0)

			gsr0 = Signal()
			gsr1 = Signal()
			# There is no end-of-startup signal on ECP5, but PUR is released after IOB enable, so
			# a simple reset synchronizer (with PUR as the asynchronous reset) does the job.
			m.submodules += [
				Instance('FD1S3AX', p_GSR = 'DISABLED', i_CK = clk_i, i_D = ~rst_i, o_Q = gsr0),
				Instance('FD1S3AX', p_GSR = 'DISABLED', i_CK = clk_i, i_D = gsr0,   o_Q = gsr1),
				# Although we already synchronize the reset input to user clock, SGSR has dedicated
				# clock routing to the center of the FPGA; use that just in case it turns out to be
				# more reliable. (None of this is documented.)
				Instance('SGSR', i_CLK = clk_i, i_GSR = gsr1),
			]
			# GSR implicitly connects to every appropriate storage element. As such, the sync
			# domain is reset-less; domains driven by other clocks would need to have dedicated
			# reset circuitry or otherwise meet setup/hold constraints on their own.
			m.domains += ClockDomain('sync', reset_less = True)
			m.d.comb += ClockSignal('sync').eq(clk_i)
			return m

	_single_ended_io_types = [
		'HSUL12', 'LVCMOS12', 'LVCMOS15', 'LVCMOS18', 'LVCMOS25', 'LVCMOS33', 'LVTTL33',
		'SSTL135_I', 'SSTL135_II', 'SSTL15_I', 'SSTL15_II', 'SSTL18_I', 'SSTL18_II',
	]
	_differential_io_types = [
		'BLVDS25', 'BLVDS25E', 'HSUL12D', 'LVCMOS18D', 'LVCMOS25D', 'LVCMOS33D',
		'LVDS', 'LVDS25E', 'LVPECL33', 'LVPECL33E', 'LVTTL33D', 'MLVDS', 'MLVDS25E',
		'SLVS', 'SSTL135D_I', 'SSTL135D_II', 'SSTL15D_I', 'SSTL15D_II', 'SSTL18D_I',
		'SSTL18D_II', 'SUBLVDS',
	]

	def should_skip_port_component(
		self, port: Subsignal, attrs: Attrs, component: Literal['io', 'i', 'o', 'p', 'n', 'oe']
	) -> bool:
		# On ECP5, a differential IO is placed by only instantiating an IO buffer primitive at
		# the PIOA or PIOC location, which is always the non-inverting pin.
		if attrs.get('IO_TYPE', 'LVCMOS25') in self._differential_io_types and component == 'n':
			return True
		return False

	def _get_xdr_buffer(
		self, m: Module, pin: Pin, *, i_invert: bool = False, o_invert: bool = False
	) -> tuple[Optional[Signal], Optional[Signal], Optional[Signal]]:
		def get_ireg(clk: Signal, d: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'IFS1P3DX',
					i_SCLK = clk,
					i_SP = Const(1),
					i_CD = Const(0),
					i_D = d[bit],
					o_Q = q[bit]
				)

		def get_oreg(clk: Signal, d: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'OFS1P3DX',
					i_SCLK = clk,
					i_SP = Const(1),
					i_CD = Const(0),
					i_D = d[bit],
					o_Q = q[bit]
				)

		def get_oereg(clk: Signal, oe: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'OFS1P3DX',
					i_SCLK = clk,
					i_SP = Const(1),
					i_CD = Const(0),
					i_D = oe,
					o_Q = q[bit]
				)

		def get_iddr(sclk: Signal, d: Signal, q0: Signal, q1: Signal) -> None:
			for bit in range(len(d)):
				m.submodules += Instance(
					'IDDRX1F',
					i_SCLK = sclk,
					i_RST = Const(0),
					i_D = d[bit],
					o_Q0 = q0[bit],
					o_Q1 = q1[bit]
				)

		def get_iddrx2(
			sclk: Signal, eclk: Signal, d: Signal,
			q0: Signal, q1: Signal, q2: Signal, q3: Signal
		) -> None:
			for bit in range(len(d)):
				m.submodules += Instance(
					'IDDRX2F',
					i_SCLK = sclk,
					i_ECLK = eclk,
					i_RST = Const(0),
					i_D = d[bit],
					o_Q0 = q0[bit],
					o_Q1 = q1[bit],
					o_Q2 = q2[bit],
					o_Q3 = q3[bit]
				)

		def get_iddr71b(
			sclk: Signal, eclk: Signal, d: Signal,
			q0: Signal, q1: Signal, q2: Signal, q3: Signal, q4: Signal, q5: Signal, q6: Signal
		) -> None:
			for bit in range(len(d)):
				m.submodules += Instance(
					'IDDR71B',
					i_SCLK = sclk,
					i_ECLK = eclk,
					i_RST = Const(0),
					i_D = d[bit],
					o_Q0 = q0[bit],
					o_Q1 = q1[bit],
					o_Q2 = q2[bit],
					o_Q3 = q3[bit],
					o_Q4 = q4[bit],
					o_Q5 = q5[bit],
					o_Q6 = q6[bit],
				)

		def get_oddr(sclk: Signal, d0: Signal, d1: Signal, q: Signal) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'ODDRX1F',
					i_SCLK = sclk,
					i_RST = Const(0),
					i_D0 = d0[bit],
					i_D1 = d1[bit],
					o_Q = q[bit]
				)

		def get_oddrx2(
			sclk: Signal, eclk: Signal,
			d0: Signal, d1: Signal, d2: Signal, d3: Signal, q: Signal
		) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'ODDRX2F',
					i_SCLK = sclk,
					i_ECLK = eclk,
					i_RST = Const(0),
					i_D0 = d0[bit],
					i_D1 = d1[bit],
					i_D2 = d2[bit],
					i_D3 = d3[bit],
					o_Q = q[bit]
				)

		def get_oddr71b(
			sclk:  Signal, eclk:  Signal,
			d0: Signal, d1: Signal, d2: Signal, d3: Signal, d4: Signal,
			d5: Signal, d6: Signal, q: Signal
		) -> None:
			for bit in range(len(q)):
				m.submodules += Instance(
					'ODDR71B',
					i_SCLK = sclk,
					i_ECLK = eclk,
					i_RST = Const(0),
					i_D0 = d0[bit],
					i_D1 = d1[bit],
					i_D2 = d2[bit],
					i_D3 = d3[bit],
					i_D4 = d4[bit],
					i_D5 = d5[bit],
					i_D6 = d6[bit],
					o_Q = q[bit]
				)

		def get_ineg(z: Signal, invert: bool) -> Signal:
			if invert:
				a = Signal.like(z, name_suffix = '_n')
				m.d.comb += z.eq(~a)
				return a
			else:
				return z

		def get_oneg(a: Signal, invert: bool) -> Signal:
			if invert:
				z = Signal.like(a, name_suffix = '_n')
				m.d.comb += z.eq(~a)
				return z
			else:
				return a

		if 'i' in pin.dir:
			if pin.xdr < 2:
				pin_i  = get_ineg(pin.i,  i_invert)
			elif pin.xdr == 2:
				pin_i0 = get_ineg(pin.i0, i_invert)
				pin_i1 = get_ineg(pin.i1, i_invert)
			elif pin.xdr == 4:
				pin_i0 = get_ineg(pin.i0, i_invert)
				pin_i1 = get_ineg(pin.i1, i_invert)
				pin_i2 = get_ineg(pin.i2, i_invert)
				pin_i3 = get_ineg(pin.i3, i_invert)
			elif pin.xdr == 7:
				pin_i0 = get_ineg(pin.i0, i_invert)
				pin_i1 = get_ineg(pin.i1, i_invert)
				pin_i2 = get_ineg(pin.i2, i_invert)
				pin_i3 = get_ineg(pin.i3, i_invert)
				pin_i4 = get_ineg(pin.i4, i_invert)
				pin_i5 = get_ineg(pin.i5, i_invert)
				pin_i6 = get_ineg(pin.i6, i_invert)
		if 'o' in pin.dir:
			if pin.xdr < 2:
				pin_o  = get_oneg(pin.o,  o_invert)
			elif pin.xdr == 2:
				pin_o0 = get_oneg(pin.o0, o_invert)
				pin_o1 = get_oneg(pin.o1, o_invert)
			elif pin.xdr == 4:
				pin_o0 = get_oneg(pin.o0, o_invert)
				pin_o1 = get_oneg(pin.o1, o_invert)
				pin_o2 = get_oneg(pin.o2, o_invert)
				pin_o3 = get_oneg(pin.o3, o_invert)
			elif pin.xdr == 7:
				pin_o0 = get_oneg(pin.o0, o_invert)
				pin_o1 = get_oneg(pin.o1, o_invert)
				pin_o2 = get_oneg(pin.o2, o_invert)
				pin_o3 = get_oneg(pin.o3, o_invert)
				pin_o4 = get_oneg(pin.o4, o_invert)
				pin_o5 = get_oneg(pin.o5, o_invert)
				pin_o6 = get_oneg(pin.o6, o_invert)

		i = o = t = None
		if 'i' in pin.dir:
			i = Signal(pin.width, name = f'{pin.name}_xdr_i')
		if 'o' in pin.dir:
			o = Signal(pin.width, name = f'{pin.name}_xdr_o')
		if pin.dir in ('oe', 'io'):
			t = Signal(pin.width, name = f'{pin.name}_xdr_t')

		if pin.xdr == 0:
			if 'i' in pin.dir:
				i = pin_i
			if 'o' in pin.dir:
				o = pin_o
			if pin.dir in ('oe', 'io'):
				t = ~pin.oe.replicate(pin.width)
		elif pin.xdr == 1:
			if 'i' in pin.dir:
				get_ireg(pin.i_clk, i, pin_i)
			if 'o' in pin.dir:
				get_oreg(pin.o_clk, pin_o, o)
			if pin.dir in ('oe', 'io'):
				get_oereg(pin.o_clk, ~pin.oe, t)
		elif pin.xdr == 2:
			if 'i' in pin.dir:
				get_iddr(pin.i_clk, i, pin_i0, pin_i1)
			if 'o' in pin.dir:
				get_oddr(pin.o_clk, pin_o0, pin_o1, o)
			if pin.dir in ('oe', 'io'):
				get_oereg(pin.o_clk, ~pin.oe, t)
		elif pin.xdr == 4:
			if 'i' in pin.dir:
				get_iddrx2(pin.i_clk, pin.i_fclk, i, pin_i0, pin_i1, pin_i2, pin_i3)
			if 'o' in pin.dir:
				get_oddrx2(pin.o_clk, pin.o_fclk, pin_o0, pin_o1, pin_o2, pin_o3, o)
			if pin.dir in ('oe', 'io'):
				get_oereg(pin.o_clk, ~pin.oe, t)
		elif pin.xdr == 7:
			if 'i' in pin.dir:
				get_iddr71b(pin.i_clk, pin.i_fclk, i, pin_i0, pin_i1, pin_i2, pin_i3, pin_i4, pin_i5, pin_i6)
			if 'o' in pin.dir:
				get_oddr71b(pin.o_clk, pin.o_fclk, pin_o0, pin_o1, pin_o2, pin_o3, pin_o4, pin_o5, pin_o6, o)
			if pin.dir in ('oe', 'io'):
				get_oereg(pin.o_clk, ~pin.oe, t)
		else:
			assert False

		return (i, o, t)

	def get_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IB',
				i_I = port.io[bit],
				o_O = i[bit]
			)
		return m

	def get_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended output', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OB',
				i_I = o[bit],
				o_O = port.io[bit]
			)
		return m

	def get_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended tristate', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBZ',
				i_T = t[bit],
				i_I = o[bit],
				o_O = port.io[bit]
			)
		return m

	def get_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input/output', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'BB',
				i_T = t[bit],
				i_I = o[bit],
				o_O = i[bit],
				io_B = port.io[bit]
			)
		return m

	def get_diff_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential input', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'IB',
				i_I = port.p[bit],
				o_O = i[bit]
			)
		return m

	def get_diff_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential output', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OB',
				i_I = o[bit],
				o_O = port.p[bit],
			)
		return m

	def get_diff_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential tristate', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'OBZ',
				i_T = t[bit],
				i_I = o[bit],
				o_O = port.p[bit],
			)
		return m

	def get_diff_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential input/output', pin, attrs, valid_xdrs = (0, 1, 2, 4, 7), valid_attrs = True
		)

		m = Module()
		i, o, t = self._get_xdr_buffer(m, pin, i_invert = invert, o_invert = invert)
		for bit in range(pin.width):
			m.submodules[f'{pin.name}_{bit}'] = Instance(
				'BB',
				i_T = t[bit],
				i_I = o[bit],
				o_O = i[bit],
				io_B = port.p[bit],
			)
		return m

	# CDC primitives are not currently specialized for ECP5.
	# While Diamond supports false path constraints; nextpnr-ecp5 does not.



# SPDX-License-Identifier: BSD-2-Clause

from collections import OrderedDict
from enum        import Enum
from functools   import reduce, wraps
from typing      import Any, Generator, Iterable, Optional, Union

from ..util      import tracer, union
from .ast        import Cat, Shape, Signal, SignalSet, Value, ValueCastable

__all__ = (
	'DIR_FANIN',
	'DIR_FANOUT',
	'DIR_NONE',
	'Direction',
	'Layout',
	'Record',
)


Direction = Enum('Direction', ('NONE', 'FANOUT', 'FANIN'))

DIR_NONE   = Direction.NONE
DIR_FANOUT = Direction.FANOUT
DIR_FANIN  = Direction.FANIN


class Layout:
	@staticmethod
	def cast(obj, *, src_loc_at: int = 0) -> 'Layout':
		if isinstance(obj, Layout):
			return obj
		return Layout(obj, src_loc_at = 1 + src_loc_at)

	# TODO: The `Any` type is not correct but the types need to be refactored again eventually to fix it
	def __init__(
		self, fields: Iterable[Union[tuple[str, Any], tuple[str, Any, Direction]]], *,
		src_loc_at: int = 0
	) -> None:
		self.fields = OrderedDict()
		for field in fields:
			if not isinstance(field, tuple) or len(field) not in (2, 3):
				raise TypeError(f'Field {field!r} has invalid layout: should be either (name, shape) or (name, shape, direction)')
			if len(field) == 2:
				name, shape = field
				direction = DIR_NONE
				if isinstance(shape, list):
					shape = Layout.cast(shape)
			else:
				name, shape, direction = field
				if not isinstance(direction, Direction):
					raise TypeError(f'Field {field!r} has invalid direction: should be a Direction instance like DIR_FANIN')
			if not isinstance(name, str):
				raise TypeError(f'Field {field!r} has invalid name: should be a string')
			if not isinstance(shape, Layout):
				try:
					# Check provided shape by calling Shape.cast and checking for exception
					Shape.cast(shape, src_loc_at = 1 + src_loc_at)
				except Exception:
					raise TypeError(
						f'Field {field!r} has invalid shape: should be castable to Shape or a list of ''fields of a nested record'
					)
			if name in self.fields:
				raise NameError(f'Field {field!r} has a name that is already present in the layout')
			self.fields[name] = (shape, direction)

	def __getitem__(self, item):
		if isinstance(item, tuple):
			return Layout([
				(name, shape, dir)
				for (name, (shape, dir)) in self.fields.items()
				if name in item
			])

		return self.fields[item]

	def __iter__(self) -> Generator[tuple[str, Any, Direction], None, None]:
		for name, (shape, dir) in self.fields.items():
			yield (name, shape, dir)

	def __eq__(self, other) -> bool:
		return self.fields == other.fields

	def __repr__(self) -> str:
		field_reprs = []
		for name, shape, dir in self:
			if dir == DIR_NONE:
				field_reprs.append(f'({name!r}, {shape!r})')
			else:
				field_reprs.append(f'({name!r}, {shape!r}, Direction.{dir.name})')
		return f'Layout([{", ".join(field_reprs)}])'


class Record(ValueCastable):
	@staticmethod
	def like(other, *, name = None, name_suffix = None, src_loc_at = 0):
		if name is not None:
			new_name = str(name)
		elif name_suffix is not None:
			new_name = other.name + str(name_suffix)
		else:
			new_name = tracer.get_var_name(depth = 2 + src_loc_at, default = None)

		def concat(a, b):
			if a is None:
				return b
			return f'{a}__{b}'

		fields = {}
		for field_name in other.fields:
			field = other[field_name]
			if isinstance(field, Record):
				fields[field_name] = Record.like(
					field, name = concat(new_name, field_name),
					src_loc_at = 1 + src_loc_at
				)
			else:
				fields[field_name] = Signal.like(
					field, name = concat(new_name, field_name),
					src_loc_at = 1 + src_loc_at
				)

		return Record(other.layout, name = new_name, fields = fields, src_loc_at = 1)

	def __init__(self, layout, *, name: Optional[str] = None, fields = None, src_loc_at: int = 0) -> None:
		if name is None:
			name = tracer.get_var_name(depth = 2 + src_loc_at, default = None)

		self.name    = name
		self.src_loc = tracer.get_src_loc(src_loc_at)

		def concat(a, b):
			if a is None:
				return b
			return f'{a}__{b}'

		self.layout = Layout.cast(layout, src_loc_at = 1 + src_loc_at)
		self.fields = OrderedDict()
		for field_name, field_shape, field_dir in self.layout:
			if fields is not None and field_name in fields:
				field = fields[field_name]
				if isinstance(field_shape, Layout):
					assert isinstance(field, Record) and field_shape == field.layout
				else:
					assert isinstance(field, Signal) and Shape.cast(field_shape) == field.shape()
				self.fields[field_name] = field
			else:
				if isinstance(field_shape, Layout):
					self.fields[field_name] = Record(
						field_shape, name = concat(name, field_name),
						src_loc_at = 1 + src_loc_at
					)
				else:
					self.fields[field_name] = Signal(
						field_shape, name = concat(name, field_name),
						src_loc_at = 1 + src_loc_at
					)

	def __getattr__(self, name):
		return self[name]

	def __getitem__(self, item):
		if isinstance(item, str):
			try:
				return self.fields[item]
			except KeyError:
				if self.name is None:
					reference = 'Unnamed record'
				else:
					reference = f'Record \'{self.name}\''
				raise AttributeError(
					f'{reference} does not have a field \'{item}\'. Did you mean one of: {", ".join(self.fields)}?'
				) from None
		elif isinstance(item, tuple):
			return Record(self.layout[item], fields = {
				field_name: field_value
				for field_name, field_value in self.fields.items()
				if field_name in item
			})
		else:
			try:
				return Value.__getitem__(self, item)
			except KeyError:
				if self.name is None:
					reference = 'Unnamed record'
				else:
					reference = f'Record \'{self.name}\''
				raise AttributeError(
					f'{reference} does not have a field \'{item}\'. Did you mean one of: {", ".join(self.fields)}?'
				) from None

	@ValueCastable.lowermethod
	def as_value(self) -> Cat:
		return Cat(self.fields.values())

	def __len__(self) -> int:
		return len(self.as_value())

	def _lhs_signals(self):
		return union((f._lhs_signals() for f in self.fields.values()), start = SignalSet())

	def _rhs_signals(self):
		return union((f._rhs_signals() for f in self.fields.values()), start = SignalSet())

	def __repr__(self) -> str:
		fields = []
		for field_name, field in self.fields.items():
			if isinstance(field, Signal):
				fields.append(field_name)
			else:
				fields.append(repr(field))
		name = self.name
		if name is None:
			name = '<unnamed>'
		return f'(rec {name} {" ".join(fields)})'

	def shape(self):
		return self.as_value().shape()

	def connect(self, *subordinates, include = None, exclude = None):
		def rec_name(record):
			if record.name is None:
				return 'unnamed record'
			else:
				return f'record \'{record.name}\''

		for field in include or {}:
			if field not in self.fields:
				raise AttributeError(f'Cannot include field \'{field}\' because it is not present in {rec_name(self)}')

		for field in exclude or {}:
			if field not in self.fields:
				raise AttributeError(f'Cannot exclude field \'{field}\' because it is not present in {rec_name(self)}')

		stmts = []
		for field in self.fields:
			if include is not None and field not in include:
				continue
			if exclude is not None and field in exclude:
				continue

			shape, direction = self.layout[field]
			if not isinstance(shape, Layout) and direction == DIR_NONE:
				raise TypeError(f'Cannot connect field \'{field}\' of {rec_name(self)} because it does not have a direction')


			item = self.fields[field]
			subord_items = []
			for subord in subordinates:
				if field not in subord.fields:
					raise AttributeError(
						f'Cannot connect field \'{field}\' of {rec_name(self)} to subordinate '
						f'{rec_name(subord)} because the subordinate record does not have this field'
					)
				subord_items.append(subord.fields[field])

			if isinstance(shape, Layout):
				sub_include = include[field] if include and field in include else None
				sub_exclude = exclude[field] if exclude and field in exclude else None
				stmts += item.connect(*subord_items, include = sub_include, exclude = sub_exclude)
			else:
				if direction == DIR_FANOUT:
					stmts += [ sub_item.eq(item) for sub_item in subord_items ]
				if direction == DIR_FANIN:
					stmts += [ item.eq(reduce(lambda a, b: a | b, subord_items)) ]

		return stmts

def _valueproxy(name):
	value_func = getattr(Value, name)

	@wraps(value_func)
	def _wrapper(self, *args, **kwargs):
		return value_func(Value.cast(self), *args, **kwargs)

	return _wrapper

for name in [
		'__bool__',
		'__invert__', '__neg__',
		'__add__', '__radd__', '__sub__', '__rsub__',
		'__mul__', '__rmul__',
		'__mod__', '__rmod__', '__floordiv__', '__rfloordiv__',
		'__lshift__', '__rlshift__', '__rshift__', '__rrshift__',
		'__and__', '__rand__', '__xor__', '__rxor__', '__or__', '__ror__',
		'__eq__', '__ne__', '__lt__', '__le__', '__gt__', '__ge__',
		'__abs__', '__len__',
		'as_unsigned', 'as_signed', 'bool', 'any', 'all', 'xor', 'implies',
		'bit_select', 'word_select', 'matches',
		'shift_left', 'shift_right', 'rotate_left', 'rotate_right', 'eq'
]:
	setattr(Record, name, _valueproxy(name))

del _valueproxy
del name



# SPDX-License-Identifier: BSD-2-Clause

from enum     import Enum, auto, unique
from typing   import Optional, Union

from ...build import Attrs, DiffPairs, Pins, PinsN, Resource, Subsignal

__all__ = (
	'QSPIMode',
	'QSPIDataMode',

	'DDR3Resource',
	'NORFlashResources',
	'QSPIFlashResource',
	'SDCardResources',
	'SDRAMResource',
	'SPIFlashResources',
	'SRAMResource',
)

@unique
class QSPIMode(Enum):
	Single       = auto()
	DualStacked  = auto()
	DualParallel = auto()

@unique
class QSPIDataMode(Enum):
	x1 = auto()
	x2 = auto()
	x4 = auto()
	x8 = auto()


def SPIFlashResources(
	*args,
	cs_n: str, clk: str, copi: str, cipo: str, wp_n: Optional[str] = None,
	hold_n: Optional[str] = None, conn: Optional[Union[tuple[str, int], int]] = None,
	attrs: Optional[Attrs] = None
) -> list[Resource]:

	resources = []

	io_all = []

	if attrs is not None:
		io_all.append(attrs)

	io_all.append(Subsignal('cs',  PinsN(cs_n, dir = 'o', conn = conn)))
	io_all.append(Subsignal('clk', Pins(clk, dir = 'o', conn = conn, assert_width = 1)))

	io_1x = list(io_all)
	io_1x.append(Subsignal('copi', Pins(copi, dir = 'o', conn = conn, assert_width = 1)))
	io_1x.append(Subsignal('cipo', Pins(cipo, dir = 'i', conn = conn, assert_width = 1)))

	if wp_n is not None and hold_n is not None:
		io_1x.append(Subsignal('wp',   PinsN(wp_n,   dir = 'o', conn = conn, assert_width = 1)))
		io_1x.append(Subsignal('hold', PinsN(hold_n, dir = 'o', conn = conn, assert_width = 1)))

	resources.append(Resource.family(
		*args, default_name = 'spi_flash', ios = io_1x, name_suffix = '1x'
	))

	io_2x = list(io_all)
	io_2x.append(Subsignal(
		'dq', Pins(' '.join([copi, cipo]), dir = 'io', conn = conn, assert_width = 2)
	))
	resources.append(Resource.family(
		*args, default_name = 'spi_flash', ios = io_2x, name_suffix = '2x'
	))

	if wp_n is not None and hold_n is not None:
		io_4x = list(io_all)
		io_4x.append(Subsignal(
			'dq', Pins(' '.join([copi, cipo, wp_n, hold_n]), dir = 'io', conn = conn, assert_width = 4)
		))
		resources.append(Resource.family(
			*args, default_name = 'spi_flash', ios = io_4x, name_suffix = '4x'
		))

	return resources

def QSPIFlashResource(
	*args, cs_n, clk, mode: QSPIMode, data_mode: QSPIDataMode,
	dq = None, dq_a = None, dq_b = None, clk_fb = None, conn = None, attrs = None
):
	if not isinstance(mode, QSPIMode) or not isinstance(data_mode, QSPIDataMode):
		raise AssertionError('mode must be a QSPIMode and data_mode a QSPIDataMode')
	elif mode == QSPIMode.Single and data_mode == QSPIDataMode.x8:
		raise AssertionError('Cannot use x8 data mode in QSPI single mode')
	elif mode == QSPIMode.DualStacked and data_mode == QSPIDataMode.x8:
		raise AssertionError('Cannot use x8 data mode in QSPI dual stacked mode')
	elif mode == QSPIMode.DualParallel and data_mode != QSPIDataMode.x8:
		raise AssertionError('Must use x8 data mode in QSPI dual parallel mode')

	clk_count = 2 if mode == QSPIMode.DualParallel else 1
	cs_count = 1 if mode == QSPIMode.Single else 2

	ios = [
		Subsignal('cs', PinsN(cs_n, dir = 'o', conn = conn, assert_width = cs_count)),
		Subsignal('clk', Pins(clk, dir = 'o', conn = conn, assert_width = clk_count)),
	]

	if mode == QSPIMode.DualParallel:
		assert dq is None and dq_a is not None and dq_b is not None
		ios.append('dq_a', Pins(dq_a, dir = 'io', conn = conn, assert_width = 4))
		ios.append('dq_b', Pins(dq_b, dir = 'io', conn = conn, assert_width = 4))
	else:
		assert dq is not None and dq_a is None and dq_b is None
		if data_mode == QSPIDataMode.x1:
			dq = dq.split(' ')
			assert len(dq) == 3
			ios.append('copi', Pins(dq[0], dir = 'o', conn = conn))
			ios.append('cipo', Pins(dq[1], dir = 'i', conn = conn))
			ios.append('hold', PinsN(dq[2], dir = 'io', conn = conn))
		elif data_mode == QSPIDataMode.x2:
			dq = dq.split(' ')
			assert len(dq) == 3
			ios.append('dq', Pins(f'{dq[0]} {dq[1]}', dir = 'io', conn = conn))
			ios.append('hold', PinsN(dq[2], dir = 'io', conn = conn))
		elif data_mode == QSPIDataMode.x4:
			ios.append('dq', Pins(dq, dir = 'io', conn = conn, assert_width = 4))

	if clk_fb is not None:
		ios.append(Subsignal('clk_fb', Pins(clk_fb, dir = 'i', conn = conn, assert_width = 1)))
	if attrs is not None:
		ios.append(attrs)
	return Resource.family(*args, ios = ios)


def SDCardResources(
	*args,
	clk: str, cmd: str, dat0: str, dat1: Optional[str] = None, dat2: Optional[str] = None,
	dat3: Optional[str] = None, cd: Optional[str] = None, wp_n: Optional[str] = None,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> list[Resource]:

	resources = []
	io_common = []

	if attrs is not None:
		io_common.append(attrs)

	if cd is not None:
		io_common.append(Subsignal('cd', Pins(cd, dir = 'i', conn = conn, assert_width = 1)))

	if wp_n is not None:
		io_common.append(Subsignal('wp', PinsN(wp_n, dir = 'i', conn = conn, assert_width = 1)))

	io_native = list(io_common)
	io_native.append(Subsignal('clk', Pins(clk, dir = 'o', conn = conn, assert_width = 1)))
	io_native.append(Subsignal('cmd', Pins(cmd, dir = 'o', conn = conn, assert_width = 1)))

	io_1bit = list(io_native)
	io_1bit.append(Subsignal('dat', Pins(dat0, dir = 'io', conn = conn, assert_width = 1)))

	if dat3 is not None:
		# DAT3 has a pullup and works as electronic card detect
		io_1bit.append(Subsignal('ecd', Pins(dat3, dir = 'i', conn = conn, assert_width = 1)))

	resources.append(Resource.family(
		*args, default_name = 'sd_card', ios = io_1bit, name_suffix = '1bit'
	))

	if dat1 is not None and dat2 is not None and dat3 is not None:
		io_4bit = list(io_native)
		io_4bit.append(Subsignal(
			'dat', Pins(' '.join((dat0, dat1, dat2, dat3)), dir = 'io', conn = conn, assert_width = 4)
		))
		resources.append(Resource.family(
			*args, default_name = 'sd_card', ios = io_4bit, name_suffix = '4bit'
		))

	if dat3 is not None:
		io_spi = list(io_common)
		# DAT3/CS# has a pullup and doubles as electronic card detect
		io_spi.append(Subsignal('cs', PinsN(dat3, dir = 'io', conn = conn, assert_width = 1)))
		io_spi.append(Subsignal('clk', Pins(clk, dir = 'o', conn = conn, assert_width = 1)))
		io_spi.append(Subsignal('copi', Pins(cmd, dir = 'o', conn = conn, assert_width = 1)))
		io_spi.append(Subsignal('cipo', Pins(dat0, dir = 'i', conn = conn, assert_width = 1)))
		resources.append(Resource.family(
			*args, default_name = 'sd_card', ios = io_spi, name_suffix = 'spi'
		))

	return resources


def SRAMResource(
	*args,
	cs_n: str, oe_n: Optional[str] = None, we_n: str, a: str, d: str,
	dm_n: Optional[str] = None, conn: Optional[Union[tuple[str, int], int]] = None,
	attrs: Optional[Attrs] = None
) -> Resource:

	io = []

	io.append(Subsignal('cs', PinsN(cs_n, dir = 'o', conn = conn, assert_width = 1)))

	if oe_n is not None:
		# Asserted WE# deactivates the D output buffers, so WE# can be used to replace OE#.
		io.append(Subsignal('oe', PinsN(oe_n, dir = 'o', conn = conn, assert_width = 1)))

	io.append(Subsignal('we', PinsN(we_n, dir = 'o', conn = conn, assert_width = 1)))
	io.append(Subsignal('a', Pins(a, dir = 'o', conn = conn)))
	io.append(Subsignal('d', Pins(d, dir = 'io', conn = conn)))

	if dm_n is not None:
		io.append(Subsignal('dm', PinsN(dm_n, dir = 'o', conn = conn))) # dm = 'LB# UB#'

	if attrs is not None:
		io.append(attrs)

	return Resource.family(*args, default_name = 'sram', ios = io)


def SDRAMResource(
	*args,
	clk: str, cke: Optional[str] = None, cs_n: Optional[str] = None,
	we_n: str, ras_n: str, cas_n: str, ba: str, a: str, dq: str,
	dqm: Optional[str] = None, conn: Optional[Union[tuple[str, int], int]] = None,
	attrs: Optional[Attrs] = None
) -> Resource:

	io = []

	io.append(Subsignal('clk', Pins(clk, dir = 'o', conn = conn, assert_width = 1)))
	if cke is not None:
		io.append(Subsignal('clk_en', Pins(cke, dir = 'o', conn = conn, assert_width = 1)))

	if cs_n is not None:
		io.append(Subsignal('cs', PinsN(cs_n,  dir = 'o', conn = conn, assert_width = 1)))

	io.append(Subsignal('we',  PinsN(we_n,  dir = 'o', conn = conn, assert_width = 1)))
	io.append(Subsignal('ras', PinsN(ras_n, dir = 'o', conn = conn, assert_width = 1)))
	io.append(Subsignal('cas', PinsN(cas_n, dir = 'o', conn = conn, assert_width = 1)))
	io.append(Subsignal('ba', Pins(ba, dir = 'o', conn = conn)))
	io.append(Subsignal('a',  Pins(a,  dir = 'o', conn = conn)))
	io.append(Subsignal('dq', Pins(dq, dir = 'io', conn = conn)))

	if dqm is not None:
		io.append(Subsignal('dqm', Pins(dqm, dir = 'o', conn = conn))) # dqm = 'LDQM# UDQM#'

	if attrs is not None:
		io.append(attrs)

	return Resource.family(*args, default_name = 'sdram', ios = io)


def NORFlashResources(
	*args,
	rst: Optional[str] = None, byte_n: Optional[str] = None,
	cs_n: str, oe_n: str, we_n: str, wp_n: str, by: str, a: str, dq: str,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> list[Resource]:

	resources = []

	io_common = []

	if rst is not None:
		io_common.append(Subsignal('rst', Pins(rst, dir = 'o', conn = conn, assert_width = 1)))

	io_common.append(Subsignal('cs', PinsN(cs_n, dir = 'o', conn = conn, assert_width = 1)))
	io_common.append(Subsignal('oe', PinsN(oe_n, dir = 'o', conn = conn, assert_width = 1)))
	io_common.append(Subsignal('we', PinsN(we_n, dir = 'o', conn = conn, assert_width = 1)))
	io_common.append(Subsignal('wp', PinsN(wp_n, dir = 'o', conn = conn, assert_width = 1)))
	io_common.append(Subsignal('rdy', Pins(by, dir = 'i', conn = conn, assert_width = 1)))

	if byte_n is None:
		io_8bit = list(io_common)
		io_8bit.append(Subsignal('a', Pins(a, dir = 'o', conn = conn)))
		io_8bit.append(Subsignal('dq', Pins(dq, dir = 'io', conn = conn, assert_width = 8)))
		resources.append(Resource.family(
			*args, default_name = 'nor_flash', ios = io_8bit, name_suffix = '8bit'
		))
	else:
		*dq_0_14, dq15_am1 = dq.split()

		# If present in a requested resource, this pin needs to be strapped correctly.
		io_common.append(Subsignal('byte', PinsN(byte_n, dir = 'o', conn = conn, assert_width = 1)))

		io_8bit = list(io_common)
		io_8bit.append(Subsignal('a', Pins(' '.join((dq15_am1, a)), dir = 'o', conn = conn)))
		io_8bit.append(Subsignal('dq', Pins(' '.join(dq_0_14[:8]), dir = 'io', conn = conn,
											assert_width = 8)))
		resources.append(Resource.family(
			*args, default_name = 'nor_flash', ios = io_8bit, name_suffix = '8bit'
		))

		io_16bit = list(io_common)
		io_16bit.append(Subsignal('a', Pins(a, dir = 'o', conn = conn)))
		io_16bit.append(Subsignal('dq', Pins(dq, dir = 'io', conn = conn, assert_width = 16)))
		resources.append(Resource.family(
			*args, default_name = 'nor_flash', ios = io_16bit, name_suffix = '16bit'
		))

	return resources


def DDR3Resource(
	*args,
	rst_n: Optional[str] = None,
	clk_p: str, clk_n: str, clk_en: str, cs_n: str, we_n: str, ras_n: str, cas_n: str,
	a: str, ba: str, dqs_p: str, dqs_n: str, dq: str, dm: str, odt: str,
	conn: Optional[Union[tuple[str, int], int]] = None,
	diff_attrs = None, attrs: Optional[Attrs] = None
) -> Resource:

	ios = []

	ios.append(Subsignal('rst', PinsN(rst_n, dir = 'o', conn = conn, assert_width = 1)))
	ios.append(Subsignal('clk', DiffPairs(clk_p, clk_n, dir = 'o', conn = conn, assert_width = 1), diff_attrs))
	ios.append(Subsignal('clk_en', Pins(clk_en, dir = 'o', conn = conn, assert_width = 1)))
	ios.append(Subsignal('cs', PinsN(cs_n, dir = 'o', conn = conn, assert_width = 1)))
	ios.append(Subsignal('we', PinsN(we_n, dir = 'o', conn = conn, assert_width = 1)))
	ios.append(Subsignal('ras', PinsN(ras_n, dir = 'o', conn = conn, assert_width = 1)))
	ios.append(Subsignal('cas', PinsN(cas_n, dir = 'o', conn = conn, assert_width = 1)))
	ios.append(Subsignal('a', Pins(a, dir = 'o', conn = conn)))
	ios.append(Subsignal('ba', Pins(ba, dir = 'o', conn = conn)))
	ios.append(Subsignal('dqs', DiffPairs(dqs_p, dqs_n, dir = 'io', conn = conn), diff_attrs))
	ios.append(Subsignal('dq', Pins(dq, dir = 'io', conn = conn)))
	ios.append(Subsignal('dm', Pins(dm, dir = 'o', conn = conn)))
	ios.append(Subsignal('odt', Pins(odt, dir = 'o', conn = conn, assert_width = 1)))

	if attrs is not None:
		ios.append(attrs)

	return Resource.family(*args, default_name = 'ddr3', ios = ios)



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl             import *
from torii.asserts         import *
from torii.sim             import *
from torii.lib.coding.gray import Encoder, Decoder

from ...utils              import ToriiTestSuiteCase

class ReversibleSpec(Elaboratable):
	def __init__(self, encoder_cls, decoder_cls, args):
		self.encoder_cls = encoder_cls
		self.decoder_cls = decoder_cls
		self.coder_args  = args

	def elaborate(self, platform):
		m = Module()
		enc, dec = self.encoder_cls(*self.coder_args), self.decoder_cls(*self.coder_args)
		m.submodules += enc, dec
		m.d.comb += [
			dec.i.eq(enc.o),
			Assert(enc.i == dec.o)
		]
		return m


class HammingDistanceSpec(Elaboratable):
	def __init__(self, distance, encoder_cls, args):
		self.distance    = distance
		self.encoder_cls = encoder_cls
		self.coder_args  = args

	def elaborate(self, platform):
		m = Module()
		enc1, enc2 = self.encoder_cls(*self.coder_args), self.encoder_cls(*self.coder_args)
		m.submodules += enc1, enc2
		m.d.comb += [
			Assume(enc1.i + 1 == enc2.i),
			Assert(sum(enc1.o ^ enc2.o) == self.distance)
		]
		return m


class GrayCoderTestCase(ToriiTestSuiteCase):
	def test_reversible(self):
		spec = ReversibleSpec(encoder_cls = Encoder, decoder_cls = Decoder, args = (16,))
		self.assertFormal(spec, mode = 'prove')

	def test_distance(self):
		spec = HammingDistanceSpec(distance = 1, encoder_cls = Encoder, args = (16,))
		self.assertFormal(spec, mode = 'prove')



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii               import *
from torii.sim           import *
from torii.lib.soc.event import *

from ...utils            import ToriiTestSuiteCase

def simulation_test(dut, process):
	sim = Simulator(dut)
	sim.add_clock(1e-6)
	sim.add_sync_process(process)
	with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
		sim.run()


class SourceTestCase(ToriiTestSuiteCase):
	def test_level(self):
		src = Source(trigger = 'level')
		self.assertEqual(src.trigger, Source.Trigger.LEVEL)

	def test_rise(self):
		src = Source(trigger = 'rise')
		self.assertEqual(src.trigger, Source.Trigger.RISE)

	def test_fall(self):
		src = Source(trigger = 'fall')
		self.assertEqual(src.trigger, Source.Trigger.FALL)

	def test_trigger_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Invalid trigger mode \'foo\'; must be one of level, rise, fall'
		):
			Source(trigger = 'foo')

	def test_get_map_wrong(self):
		src = Source()
		with self.assertRaisesRegex(
			NotImplementedError,
			r'Event source \(rec src i trg\) does not have an event map'
		):
			src.event_map

	def test_get_map_frozen(self):
		src = Source()
		src.event_map = EventMap()
		with self.assertRaisesRegex(
			ValueError,
			r'Event map has been frozen. Cannot add source.'
		):
			src.event_map.add(Source())

	def test_set_map_wrong(self):
		src = Source()
		with self.assertRaisesRegex(
			TypeError,
			r'Event map must be an instance of EventMap, not \'foo\''
		):
			src.event_map = 'foo'


class EventMapTestCase(ToriiTestSuiteCase):
	def test_add(self):
		src_0 = Source()
		src_1 = Source()
		event_map = EventMap()
		event_map.add(src_0)
		event_map.add(src = src_1)
		self.assertTrue(src_0 in event_map._sources)
		self.assertTrue(src_1 in event_map._sources)

	def test_add_wrong(self):
		event_map = EventMap()
		with self.assertRaisesRegex(
			TypeError,
			r'Event source must be an instance of event.Source, not \'foo\''
		):
			event_map.add('foo')

	def test_add_wrong_frozen(self):
		event_map = EventMap()
		event_map.freeze()
		with self.assertRaisesRegex(
			ValueError,
			r'Event map has been frozen. Cannot add source.'
		):
			event_map.add(Source())

	def test_size(self):
		event_map = EventMap()
		event_map.add(Source())
		event_map.add(Source())
		self.assertEqual(event_map.size, 2)

	def test_index(self):
		src_0 = Source()
		src_1 = Source()
		event_map = EventMap()
		event_map.add(src_0)
		event_map.add(src_1)
		self.assertEqual(event_map.index(src_0), 0)
		self.assertEqual(event_map.index(src = src_1), 1)

	def test_index_add_twice(self):
		src = Source()
		event_map = EventMap()
		event_map.add(src)
		event_map.add(src)
		self.assertEqual(event_map.index(src), 0)
		self.assertEqual(event_map.size, 1)

	def test_index_wrong(self):
		event_map = EventMap()
		with self.assertRaisesRegex(
			TypeError,
			r'Event source must be an instance of event.Source, not \'foo\''
		):
			event_map.index('foo')

	def test_index_not_found(self):
		src = Source()
		event_map = EventMap()
		with self.assertRaises(KeyError):
			event_map.index(src)

	def test_iter_sources(self):
		src_0 = Source()
		src_1 = Source()
		event_map = EventMap()
		event_map.add(src_0)
		event_map.add(src_1)
		self.assertEqual(list(event_map.sources()), [
			(src_0, 0),
			(src_1, 1),
		])


class MonitorTestCase(ToriiTestSuiteCase):
	def test_simple(self):
		sub_0 = Source()
		sub_1 = Source()
		event_map = EventMap()
		event_map.add(sub_0)
		event_map.add(sub_1)
		dut = Monitor(event_map, trigger = 'rise')
		self.assertIs(dut.src.event_map, event_map)
		self.assertEqual(dut.src.trigger, Source.Trigger.RISE)
		self.assertEqual(dut.enable.width, 2)
		self.assertEqual(dut.pending.width, 2)
		self.assertEqual(dut.clear.width, 2)

	def test_event_map_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Event map must be an instance of EventMap, not \'foo\''
		):
			Monitor(event_map = 'foo')

	def test_events(self):
		sub_0 = Source(trigger = 'level')
		sub_1 = Source(trigger = 'rise')
		sub_2 = Source(trigger = 'fall')
		event_map = EventMap()
		event_map.add(sub_0)
		event_map.add(sub_1)
		event_map.add(sub_2)
		dut = Monitor(event_map)

		def process():
			yield sub_0.i.eq(1)
			yield sub_1.i.eq(0)
			yield sub_2.i.eq(1)
			yield
			self.assertEqual((yield sub_0.trg), 1)
			self.assertEqual((yield sub_1.trg), 0)
			self.assertEqual((yield sub_2.trg), 0)
			yield
			self.assertEqual((yield dut.pending), 0b001)
			self.assertEqual((yield dut.src.i), 0)

			yield dut.enable.eq(0b111)
			yield
			self.assertEqual((yield dut.src.i), 1)

			yield dut.clear.eq(0b001)
			yield
			self.assertEqual((yield dut.pending), 0b001)
			self.assertEqual((yield dut.src.i), 1)

			yield sub_0.i.eq(0)
			yield
			self.assertEqual((yield sub_0.trg), 0)
			self.assertEqual((yield sub_1.trg), 0)
			self.assertEqual((yield sub_2.trg), 0)
			yield
			self.assertEqual((yield dut.pending), 0b000)
			self.assertEqual((yield dut.src.i), 0)

			yield sub_1.i.eq(1)
			yield
			self.assertEqual((yield sub_0.trg), 0)
			self.assertEqual((yield sub_1.trg), 1)
			self.assertEqual((yield sub_2.trg), 0)
			yield
			self.assertEqual((yield dut.pending), 0b010)
			self.assertEqual((yield dut.src.i), 1)

			yield sub_2.i.eq(0)
			yield
			self.assertEqual((yield sub_0.trg), 0)
			self.assertEqual((yield sub_1.trg), 0)
			self.assertEqual((yield sub_2.trg), 1)
			yield
			self.assertEqual((yield dut.pending), 0b110)
			self.assertEqual((yield dut.src.i), 1)

			yield dut.clear.eq(0b110)
			yield
			self.assertEqual((yield sub_0.trg), 0)
			self.assertEqual((yield sub_1.trg), 0)
			self.assertEqual((yield sub_2.trg), 0)
			yield
			self.assertEqual((yield dut.pending), 0b000)
			self.assertEqual((yield dut.src.i), 0)

		simulation_test(dut, process)



# SPDX-License-Identifier: BSD-2-Clause
# If the design does not create a "sync" clock domain, it is created by the Torii build system
# using the platform default clock (and default reset, if any).

from torii                                     import Elaboratable, Signal, Module
from torii_boards.lattice.ice40_hx1k_blink_evn import ICE40HX1KBlinkEVNPlatform

class Blinky(Elaboratable):
	def elaborate(self, platform) -> Module:
		led   = platform.request('led', 0)
		timer = Signal(20)

		m = Module()
		m.d.sync += timer.eq(timer + 1)
		m.d.comb += led.o.eq(timer[-1])
		return m


if __name__ == '__main__':
	platform = ICE40HX1KBlinkEVNPlatform()
	platform.build(Blinky(), do_program = True)



# SPDX-License-Identifier: BSD-2-Clause

from .cxxrtl  import convert          as cxxrtl_convert
from .cxxrtl  import convert_fragment as cxxrtl_convert_fragment

from .rtlil   import convert          as rtlil_convert
from .rtlil   import convert_fragment as rtlil_convert_fragment

from .verilog import convert          as verilog_convert
from .verilog import convert_fragment as verilog_convert_fragment

__all__ = (
	'cxxrtl_convert',
	'cxxrtl_convert_fragment',

	'rtlil_convert',
	'rtlil_convert_fragment',

	'verilog_convert',
	'verilog_convert_fragment',
)



# SPDX-License-Identifier: BSD-2-Clause

import io
import re
from collections  import OrderedDict
from contextlib   import contextmanager
from typing       import Literal, Optional, Union

from ..hdl        import ast, ir, mem, xfrm
from ..util       import flatten
from ..util.units import bits_for

__all__ = (
	'convert_fragment',
	'convert',
)

_escape_map = str.maketrans({
	'\"': '\\\"',
	'\\': '\\\\',
	'\t': '\\t',
	'\r': '\\r',
	'\n': '\\n',
})


def _signed(value: Union[str, int, ast.Const]) -> bool:
	if isinstance(value, str):
		return False
	elif isinstance(value, int):
		return value < 0
	elif isinstance(value, ast.Const):
		return value.signed
	else:
		raise TypeError(f'Expected one of \'str\', \'int\', \'Const\', not {value!r}')


def _const(value: Union[str, int, ast.Const]) -> str:
	if isinstance(value, str):
		return f'\"{value.translate(_escape_map)}\"'
	elif isinstance(value, int):
		if value in range(0, 2**31-1):
			return f'{value:d}'
		else:
			# This code path is only used for Instances, where Verilog-like behavior is desirable.
			# Verilog ensures that integers with unspecified width are 32 bits wide or more.
			width = max(32, bits_for(value))
			return _const(ast.Const(value, width))
	elif isinstance(value, ast.Const):
		value_twos_compl = value.value & ((1 << value.width) - 1)
		return f'{value.width}\'{value_twos_compl:0{value.width}b}'
	else:
		raise TypeError(f'Expected one of \'str\', \'int\', \'Const\', not {value!r}')


class _Namer:
	def __init__(self) -> None:
		super().__init__()
		self._anon  = 0
		self._index = 0
		self._names = set()

	def anonymous(self) -> str:
		name = f'U$${self._anon}'
		assert name not in self._names
		self._anon += 1
		return name

	def _make_name(self, name: Optional[str], local: bool) -> str:
		if name is None:
			self._index += 1
			name = f'${self._index}'
		elif not local and name[0] not in '\\$':
			name = f'\\{name}'
		while name in self._names:
			self._index += 1
			name = f'{name}${self._index}'
		self._names.add(name)
		return name


class _BufferedBuilder:
	def __init__(self) -> None:
		super().__init__()
		self._buffer = io.StringIO()

	def __str__(self) -> str:
		return self._buffer.getvalue()

	def _append(self, fmt: str, *args, **kwargs) -> None:
		self._buffer.write(fmt.format(*args, **kwargs))


class _ProxiedBuilder:
	def _append(self, *args, **kwargs) -> None:
		self.rtlil._append(*args, **kwargs)


class _AttrBuilder:
	def __init__(self, emit_src: bool, *args, **kwargs) -> None:
		super().__init__(*args, **kwargs)
		self.emit_src = emit_src

	def _attribute(self, name: str, value: Union[str, int, ast.Const], *, indent: int = 0) -> None:
		self._append('{}attribute \\{} {}\n', '  ' * indent, name, _const(value))

	def _attributes(self, attrs: dict[str, Union[str, int, ast.Const]], *, src: str = None, **kwargs) -> None:
		for name, value in attrs.items():
			self._attribute(name, value, **kwargs)
		if src and self.emit_src:
			self._attribute('src', src, **kwargs)

class _Builder(_BufferedBuilder, _Namer):
	def __init__(self, emit_src: bool) -> None:
		super().__init__()
		self.emit_src = emit_src

	def module(self, name: str = None, attrs: dict[str, Union[str, int, ast.Const]] = {}) -> '_ModuleBuilder':
		name = self._make_name(name, local = False)
		return _ModuleBuilder(self, name, attrs)

class _ModuleBuilder(_AttrBuilder, _BufferedBuilder, _Namer):
	def __init__(self, rtlil, name: str, attrs: dict[str, Union[str, int, ast.Const]]) -> None:
		super().__init__(emit_src = rtlil.emit_src)
		self.rtlil = rtlil
		self.name  = name
		self.attrs = { 'generator': 'Torii' }
		self.attrs.update(attrs)

	def __enter__(self) -> '_ModuleBuilder':
		self._attributes(self.attrs)
		self._append('module {}\n', self.name)
		return self

	def __exit__(self, *args) -> None:
		self._append('end\n')
		self.rtlil._buffer.write(str(self))

	# FIXME: Type annotation for `port_id` seems to be incorrect
	def wire(
		self, width: int, port_id: Optional[str] = None, port_kind: Literal['input', 'output', 'inout'] = None,
		name: str = None, attrs: dict[str, Union[str, int, ast.Const]] = {}, src: str = ''
	) -> str:
		# Very large wires are unlikely to work. Verilog 1364-2005 requires the limit on vectors
		# to be at least 2**16 bits, and Yosys 0.9 cannot read RTLIL with wires larger than 2**32
		# bits. In practice, wires larger than 2**16 bits, although accepted, cause performance
		# problems without an immediately visible cause, so conservatively limit wire size.
		if width > 2 ** 16:
			raise OverflowError(
				f'Wire created at {src or "unknown location"} is {width} bits wide, which is unlikely to '
				'synthesize correctly'
			)

		self._attributes(attrs, src = src, indent = 1)
		name = self._make_name(name, local = False)
		if port_id is None:
			self._append('  wire width {} {}\n', width, name)
		else:
			if port_kind not in ('input', 'output', 'inout'):
				raise ValueError(f'Expected one of \'input\', \'output\', \'inout\' for port_kind, not {port_kind!r}')
			# By convention, Yosys ports named $\d+ are positional, so there is no way to use
			# a port with such a name. See amaranth-lang/amaranth#733.
			if port_id is None:
				# NOTE: This seems like it's not purely needed? as to get to this branch we ensure
				# that `port_id` is not None anyway.
				raise ValueError(
					'Yosys ports named \'$\\d+\' are positional, and Torii does not support '
					'connecting cell ports py position.'
				)
			self._append('  wire width {} {} {} {}\n', width, port_kind, port_id, name)
		return name

	def connect(self, lhs, rhs) -> None:
		self._append('  connect {} {}\n', lhs, rhs)

	def memory(
		self, width: int, size: int, name: str = None, attrs: dict[str, Union[str, int, ast.Const]] = {}, src: str = ''
	) -> str:
		self._attributes(attrs, src = src, indent = 1)
		name = self._make_name(name, local = False)
		self._append('  memory width {} size {} {}\n', width, size, name)
		return name

	def cell(
		self, kind: str, name: str = None, params: dict[str, Union[str, int, float, ast.Const]] = {},
		ports: dict[str, str] = {}, attrs: dict[str, Union[str, int, ast.Const]] = {}, src: str = ''
	) -> str:
		self._attributes(attrs, src = src, indent = 1)
		name = self._make_name(name, local = False)
		self._append('  cell {} {}\n', kind, name)
		for param, value in params.items():
			if isinstance(value, float):
				self._append('    parameter real \\{} \"{!r}\"\n', param, value)
			elif _signed(value):
				self._append('    parameter signed \\{} {}\n', param, _const(value))
			else:
				self._append('    parameter \\{} {}\n', param, _const(value))
		for port, wire in ports.items():
			# By convention, Yosys ports named $\d+ are positional. Torii does not support
			# connecting cell ports by position. See amaranth-lang/amaranth#733.
			if re.match(r'^\$\d+$', port):
				raise ValueError(
					f'Invalid port name \'{port}\'.\n Yosys ports named \'$\\d+\' are positional,'
					'and Torii does not support connecting cell ports py position.'
				)
			self._append('    connect {} {}\n', port, wire)
		self._append('  end\n')
		return name

	def process(
		self, name: Optional[str] = None, attrs: dict[str, Union[str, int, ast.Const]] = {}, src: str = ''
	) -> '_ProcessBuilder':
		name = self._make_name(name, local = True)
		return _ProcessBuilder(self, name, attrs, src)

class _ProcessBuilder(_AttrBuilder, _BufferedBuilder):
	def __init__(self, rtlil, name: str , attrs: dict[str, Union[str, int, ast.Const]], src: str) -> None:
		super().__init__(emit_src = rtlil.emit_src)
		self.rtlil = rtlil
		self.name  = name
		self.attrs = {}
		self.src   = src

	def __enter__(self) -> '_ProcessBuilder':
		self._attributes(self.attrs, src = self.src, indent = 1)
		self._append('  process {}\n', self.name)
		return self

	def __exit__(self, *args) -> None:
		self._append('  end\n')
		self.rtlil._buffer.write(str(self))

	def case(self) -> '_CaseBuilder':
		return _CaseBuilder(self, indent = 2)

	def sync(self, kind: Literal['i', 'o', 'io'], cond: Optional[str] = None) -> '_SyncBuilder':
		return _SyncBuilder(self, kind, cond)


class _CaseBuilder(_ProxiedBuilder):
	def __init__(self, rtlil, indent: int):
		self.rtlil  = rtlil
		self.indent = indent

	def __enter__(self) -> '_CaseBuilder':
		return self

	def __exit__(self, *args) -> None:
		pass

	def assign(self, lhs, rhs):
		self._append('{}assign {} {}\n', '  ' * self.indent, lhs, rhs)

	def switch(
		self, cond, attrs: dict[str, Union[str, int, ast.Const]] = {}, src: str = ''
	) -> '_SwitchBuilder':
		return _SwitchBuilder(self.rtlil, cond, attrs, src, self.indent)


class _SwitchBuilder(_AttrBuilder, _ProxiedBuilder):
	def __init__(
		self, rtlil, cond, attrs: dict[str, Union[str, int, ast.Const]] , src: str, indent: int
	) -> None:
		super().__init__(emit_src = rtlil.emit_src)
		self.rtlil  = rtlil
		self.cond   = cond
		self.attrs  = attrs
		self.src    = src
		self.indent = indent

	def __enter__(self) -> '_SwitchBuilder':
		self._attributes(self.attrs, src = self.src, indent = self.indent)
		self._append('{}switch {}\n', '  ' * self.indent, self.cond)
		return self

	def __exit__(self, *args) -> None:
		self._append('{}end\n', '  ' * self.indent)

	def case(
		self, *values, attrs: dict[str, Union[str, int, ast.Const]] = {}, src: str = ''
	) -> '_CaseBuilder':
		self._attributes(attrs, src = src, indent = self.indent + 1)
		if values == ():
			self._append('{}case\n', '  ' * (self.indent + 1))
		else:
			self._append(
				'{}case {}\n', '  ' * (self.indent + 1), ', '.join(f'{len(value)}\'{value}'for value in values)
			)
		return _CaseBuilder(self.rtlil, self.indent + 2)


class _SyncBuilder(_ProxiedBuilder):
	def __init__(self, rtlil, kind: Literal['i', 'o', 'io'], cond: Optional[str]) -> None:
		self.rtlil = rtlil
		self.kind  = kind
		self.cond  = cond

	def __enter__(self) -> '_SyncBuilder':
		if self.cond is None:
			self._append('    sync {}\n', self.kind)
		else:
			self._append('    sync {} {}\n', self.kind, self.cond)
		return self

	def __exit__(self, *args) -> None:
		pass

	def update(self, lhs, rhs) -> None:
		self._append('      update {} {}\n', lhs, rhs)


def _src(src_loc: Optional[tuple[str, int]]) -> Optional[str]:
	if src_loc is None:
		return None
	file, line = src_loc
	return f'{file}:{line}'


class _LegalizeValue(Exception):
	def __init__(self, value, branches, src_loc: Optional[tuple[str, int]]) -> None:
		self.value    = value
		self.branches = list(branches)
		self.src_loc  = src_loc


class _ValueCompilerState:
	def __init__(self, rtlil) -> None:
		self.rtlil  = rtlil
		self.wires  = ast.SignalDict()
		self.driven = ast.SignalDict()
		self.ports  = ast.SignalDict()
		self.anys   = ast.ValueDict()

		self.expansions = ast.ValueDict()

	def add_driven(self, signal, sync) -> None:
		self.driven[signal] = sync

	def add_port(self, signal, kind: Literal['i', 'o', 'io']) -> None:
		if kind not in ('i', 'o', 'io'):
			raise ValueError(f'Expected one of \'i\', \'o\', \'io\' not {kind!r}')
		if kind == 'i':
			kind = 'input'
		elif kind == 'o':
			kind = 'output'
		elif kind == 'io':
			kind = 'inout'
		self.ports[signal] = (len(self.ports), kind)

	def resolve(self, signal, prefix = None):
		if len(signal) == 0:
			return '{ }', '{ }'

		if signal in self.wires:
			return self.wires[signal]

		if signal in self.ports:
			port_id, port_kind = self.ports[signal]
		else:
			port_id = port_kind = None
		if prefix is not None:
			wire_name = f'{prefix}_{signal.name}'
		else:
			wire_name = signal.name

		is_sync_driven = signal in self.driven and self.driven[signal]

		attrs = dict(signal.attrs)
		if signal._enum_class is not None:
			attrs['enum_base_type'] = signal._enum_class.__name__
			for value in signal._enum_class:
				attrs[f'enum_value_{value.value:0{signal.width}b}'] = value.name

		# For every signal in the sync domain, assign \sig's initial value (using the \init reg
		# attribute) to the reset value.
		if is_sync_driven:
			attrs['init'] = ast.Const(signal.reset, signal.width)

		wire_curr = self.rtlil.wire(width = signal.width, name = wire_name,
									port_id = port_id, port_kind = port_kind,
									attrs = attrs, src = _src(signal.src_loc))
		if is_sync_driven:
			wire_next = self.rtlil.wire(width = signal.width, name = wire_curr + '$next',
										src = _src(signal.src_loc))
		else:
			wire_next = None
		self.wires[signal] = (wire_curr, wire_next)

		return (wire_curr, wire_next)

	def resolve_curr(self, signal, prefix = None):
		wire_curr, wire_next = self.resolve(signal, prefix)
		return wire_curr

	def expand(self, value):
		if not self.expansions:
			return value
		return self.expansions.get(value, value)

	@contextmanager
	def expand_to(self, value, expansion):
		try:
			assert value not in self.expansions
			self.expansions[value] = expansion
			yield
		finally:
			del self.expansions[value]


class _ValueCompiler(xfrm.ValueVisitor):
	def __init__(self, state):
		self.s = state

	def on_unknown(self, value):
		if value is None:
			return None
		else:
			super().on_unknown(value)

	def on_ClockSignal(self, value):
		raise NotImplementedError # :nocov:

	def on_ResetSignal(self, value):
		raise NotImplementedError # :nocov:

	def on_Sample(self, value):
		raise NotImplementedError # :nocov:

	def on_Initial(self, value):
		raise NotImplementedError # :nocov:

	def on_Cat(self, value):
		return f'{{ {" ".join(reversed([self(o) for o in value.parts]))} }}'

	def _prepare_value_for_Slice(self, value):
		raise NotImplementedError # :nocov:

	def on_Slice(self, value):
		if value.start == 0 and value.stop == len(value.value):
			return self(value.value)

		sigspec = self._prepare_value_for_Slice(value.value)

		if value.start == value.stop:
			return '{}'
		elif value.start + 1 == value.stop:
			return f'{sigspec} [{value.start}]'
		else:
			return f'{sigspec} [{value.stop - 1}:{value.start}]'

	def on_ArrayProxy(self, value):
		index = self.s.expand(value.index)
		if isinstance(index, ast.Const):
			if index.value < len(value.elems):
				elem = value.elems[index.value]
			else:
				elem = value.elems[-1]
			return self.match_shape(elem, value.shape())
		else:
			max_index = 1 << len(value.index)
			max_elem  = len(value.elems)
			raise _LegalizeValue(value.index, range(min(max_index, max_elem)), value.src_loc)


class _RHSValueCompiler(_ValueCompiler):
	operator_map = {
		(1, '~') : '$not',
		(1, '-') : '$neg',
		(1, 'b') : '$reduce_bool',
		(1, 'r|'): '$reduce_or',
		(1, 'r&'): '$reduce_and',
		(1, 'r^'): '$reduce_xor',
		(2, '+') : '$add',
		(2, '-') : '$sub',
		(2, '*') : '$mul',
		(2, '//'): '$divfloor',
		(2, '%') : '$modfloor',
		(2, '**'): '$pow',
		(2, '<<'): '$sshl',
		(2, '>>'): '$sshr',
		(2, '&') : '$and',
		(2, '^') : '$xor',
		(2, '|') : '$or',
		(2, '=='): '$eq',
		(2, '!='): '$ne',
		(2, '<') : '$lt',
		(2, '<='): '$le',
		(2, '>') : '$gt',
		(2, '>='): '$ge',
		(3, 'm') : '$mux',
	}

	def on_value(self, value):
		return super().on_value(self.s.expand(value))

	def on_Const(self, value):
		return _const(value)

	def on_AnyConst(self, value):
		if value in self.s.anys:
			return self.s.anys[value]

		res_shape = value.shape()
		res = self.s.rtlil.wire(width = res_shape.width, src = _src(value.src_loc))
		self.s.rtlil.cell('$anyconst', ports = {
			'\\Y': res,
		}, params = {
			'WIDTH': res_shape.width,
		}, src = _src(value.src_loc))
		self.s.anys[value] = res
		return res

	def on_AnySeq(self, value):
		if value in self.s.anys:
			return self.s.anys[value]

		res_shape = value.shape()
		res = self.s.rtlil.wire(width = res_shape.width, src = _src(value.src_loc))
		self.s.rtlil.cell('$anyseq', ports = {
			'\\Y': res,
		}, params = {
			'WIDTH': res_shape.width,
		}, src = _src(value.src_loc))
		self.s.anys[value] = res
		return res

	def on_Signal(self, value):
		wire_curr, wire_next = self.s.resolve(value)
		return wire_curr

	def on_Operator_unary(self, value):
		arg, = value.operands
		if value.operator in ('u', 's'):
			# These operators don't change the bit pattern, only its interpretation.
			return self(arg)

		arg_shape = arg.shape()
		res_shape = value.shape()
		res = self.s.rtlil.wire(width = res_shape.width, src = _src(value.src_loc))
		self.s.rtlil.cell(self.operator_map[(1, value.operator)], ports = {
			'\\A': self(arg),
			'\\Y': res,
		}, params = {
			'A_SIGNED': arg_shape.signed,
			'A_WIDTH' : arg_shape.width,
			'Y_WIDTH' : res_shape.width,
		}, src = _src(value.src_loc))
		return res

	def match_shape(self, value, new_shape):
		if isinstance(value, ast.Const):
			return self(ast.Const(value.value, new_shape))

		value_shape = value.shape()
		if new_shape.width <= value_shape.width:
			return self(ast.Slice(value, 0, new_shape.width))

		res = self.s.rtlil.wire(width = new_shape.width, src = _src(value.src_loc))
		self.s.rtlil.cell('$pos', ports = {
			'\\A': self(value),
			'\\Y': res,
		}, params = {
			'A_SIGNED': value_shape.signed,
			'A_WIDTH' : value_shape.width,
			'Y_WIDTH' : new_shape.width,
		}, src = _src(value.src_loc))
		return res

	def on_Operator_binary(self, value):
		lhs, rhs = value.operands
		lhs_shape = lhs.shape()
		rhs_shape = rhs.shape()
		res_shape = value.shape()

		if lhs_shape.signed == rhs_shape.signed or value.operator in ('<<', '>>', '**'):
			lhs_wire = self(lhs)
			rhs_wire = self(rhs)
		else:
			lhs_shape = rhs_shape = ast.signed(max(lhs_shape.width + rhs_shape.signed, rhs_shape.width + lhs_shape.signed))

			lhs_wire = self.match_shape(lhs, lhs_shape)
			rhs_wire = self.match_shape(rhs, rhs_shape)
		res = self.s.rtlil.wire(width = res_shape.width, src = _src(value.src_loc))
		self.s.rtlil.cell(self.operator_map[(2, value.operator)], ports = {
			'\\A': lhs_wire,
			'\\B': rhs_wire,
			'\\Y': res,
		}, params = {
			'A_SIGNED': lhs_shape.signed,
			'A_WIDTH' : lhs_shape.width,
			'B_SIGNED': rhs_shape.signed,
			'B_WIDTH' : rhs_shape.width,
			'Y_WIDTH' : res_shape.width,
		}, src = _src(value.src_loc))
		if value.operator in ('//', '%'):
			# RTLIL leaves division by zero undefined, but we require it to return zero.
			divmod_res = res
			res = self.s.rtlil.wire(width = res_shape.width, src = _src(value.src_loc))
			self.s.rtlil.cell('$mux', ports = {
				'\\A': divmod_res,
				'\\B': self(ast.Const(0, res_shape)),
				'\\S': self(rhs == 0),
				'\\Y': res,
			}, params = {
				'WIDTH': res_shape.width
			}, src = _src(value.src_loc))
		return res

	def on_Operator_mux(self, value):
		sel, val1, val0 = value.operands
		if len(sel) != 1:
			sel = sel.bool()
		res_shape = value.shape()
		val1_wire = self.match_shape(val1, res_shape)
		val0_wire = self.match_shape(val0, res_shape)
		res = self.s.rtlil.wire(width = res_shape.width, src = _src(value.src_loc))
		self.s.rtlil.cell('$mux', ports = {
			'\\A': val0_wire,
			'\\B': val1_wire,
			'\\S': self(sel),
			'\\Y': res,
		}, params = {
			'WIDTH': res_shape.width
		}, src = _src(value.src_loc))
		return res

	def on_Operator(self, value):
		if len(value.operands) == 1:
			return self.on_Operator_unary(value)
		elif len(value.operands) == 2:
			return self.on_Operator_binary(value)
		elif len(value.operands) == 3:
			assert value.operator == 'm'
			return self.on_Operator_mux(value)
		else:
			raise TypeError # :nocov:

	def _prepare_value_for_Slice(self, value):
		if isinstance(value, (ast.Signal, ast.Slice, ast.Cat)):
			sigspec = self(value)
		else:
			sigspec = self.s.rtlil.wire(len(value), src = _src(value.src_loc))
			self.s.rtlil.connect(sigspec, self(value))
		return sigspec

	def on_Part(self, value):
		lhs, rhs = value.value, value.offset
		if value.stride != 1:
			rhs *= value.stride
		lhs_shape = lhs.shape()
		rhs_shape = rhs.shape()
		res_shape = value.shape()
		res = self.s.rtlil.wire(width = res_shape.width, src = _src(value.src_loc))
		# Note: Verilog's x[o+:w] construct produces a $shiftx cell, not a $shift cell.
		# However, Torii's semantics defines the out-of-range bits to be zero, so it is correct
		# to use a $shift cell here instead, even though it produces less idiomatic Verilog.
		self.s.rtlil.cell('$shift', ports = {
			'\\A': self(lhs),
			'\\B': self(rhs),
			'\\Y': res,
		}, params = {
			'A_SIGNED': lhs_shape.signed,
			'A_WIDTH' : lhs_shape.width,
			'B_SIGNED': rhs_shape.signed,
			'B_WIDTH' : rhs_shape.width,
			'Y_WIDTH' : res_shape.width,
		}, src = _src(value.src_loc))
		return res


class _LHSValueCompiler(_ValueCompiler):
	def on_Const(self, value):
		raise TypeError # :nocov:

	def on_AnyConst(self, value):
		raise TypeError # :nocov:

	def on_AnySeq(self, value):
		raise TypeError # :nocov:

	def on_Operator(self, value):
		if value.operator in ('u', 's'):
			# These operators are transparent on the LHS.
			arg, = value.operands
			return self(arg)

		raise TypeError # :nocov:

	def match_shape(self, value, new_shape):
		value_shape = value.shape()
		if new_shape.width == value_shape.width:
			return self(value)
		elif new_shape.width < value_shape.width:
			return self(ast.Slice(value, 0, new_shape.width))
		else: # new_shape.width > value_shape.width
			dummy_bits = new_shape.width - value_shape.width
			dummy_wire = self.s.rtlil.wire(dummy_bits)
			return f'{{ {dummy_wire} {self(value)} }}'

	def on_Signal(self, value):
		if value not in self.s.driven:
			raise ValueError(f'No LHS wire for non-driven signal {repr(value)}')
		wire_curr, wire_next = self.s.resolve(value)
		return wire_next or wire_curr

	def _prepare_value_for_Slice(self, value):
		assert isinstance(value, (ast.Signal, ast.Slice, ast.Cat, ast.Part))
		return self(value)

	def on_Part(self, value):
		offset = self.s.expand(value.offset)
		if isinstance(offset, ast.Const):
			start = offset.value * value.stride
			stop  = start + value.width
			slice = self(ast.Slice(value.value, start, min(len(value.value), stop)))
			if len(value.value) >= stop:
				return slice
			else:
				dummy_wire = self.s.rtlil.wire(stop - len(value.value))
				return f'{{ {dummy_wire} {slice} }}'
		else:
			# Only so many possible parts. The amount of branches is exponential; if value.offset
			# is large (e.g. 32-bit wide), trying to naively legalize it is likely to exhaust
			# system resources.
			max_branches = len(value.value) // value.stride + 1
			raise _LegalizeValue(
				value.offset,
				range(1 << len(value.offset))[:max_branches],
				value.src_loc
			)

class _StatementCompiler(xfrm.StatementVisitor):
	def __init__(self, state, rhs_compiler, lhs_compiler):
		self.state        = state
		self.rhs_compiler = rhs_compiler
		self.lhs_compiler = lhs_compiler

		self._case        = None
		self._test_cache  = {}
		self._has_rhs     = False
		self._wrap_assign = False

	@contextmanager
	def case(self, switch, values, attrs = {}, src = ''):
		try:
			old_case = self._case
			with switch.case(*values, attrs = attrs, src = src) as self._case:
				yield
		finally:
			self._case = old_case

	def _check_rhs(self, value):
		if self._has_rhs or next(iter(value._rhs_signals()), None) is not None:
			self._has_rhs = True

	def on_Assign(self, stmt):
		self._check_rhs(stmt.rhs)

		lhs_shape = stmt.lhs.shape()
		rhs_shape = stmt.rhs.shape()
		if lhs_shape.width == rhs_shape.width:
			rhs_sigspec = self.rhs_compiler(stmt.rhs)
		else:
			# In RTLIL, LHS and RHS of assignment must have exactly same width.
			rhs_sigspec = self.rhs_compiler.match_shape(stmt.rhs, lhs_shape)
		if self._wrap_assign:
			# In RTLIL, all assigns are logically sequenced before all switches, even if they are
			# interleaved in the source. In Torii, the source ordering is used. To handle this
			# mismatch, we wrap all assigns following a switch in a dummy switch.
			with self._case.switch('{ }') as wrap_switch:
				with wrap_switch.case() as wrap_case:
					wrap_case.assign(self.lhs_compiler(stmt.lhs), rhs_sigspec)
		else:
			self._case.assign(self.lhs_compiler(stmt.lhs), rhs_sigspec)

	def on_property(self, stmt):
		self(stmt._check.eq(stmt.test))
		self(stmt._en.eq(1))

		en_wire = self.rhs_compiler(stmt._en)
		check_wire = self.rhs_compiler(stmt._check)
		self.state.rtlil.cell('$' + stmt._kind, ports = {
			'\\A' : check_wire,
			'\\EN': en_wire,
		}, src = _src(stmt.src_loc), name = stmt.name)

	on_Assert = on_property
	on_Assume = on_property
	on_Cover  = on_property

	def on_Switch(self, stmt):
		self._check_rhs(stmt.test)

		if not self.state.expansions:
			# We repeatedly translate the same switches over and over (see the LHSGroupAnalyzer
			# related code below), and translating the switch test only once helps readability.
			if stmt not in self._test_cache:
				self._test_cache[stmt] = self.rhs_compiler(stmt.test)
			test_sigspec = self._test_cache[stmt]
		else:
			# However, if the switch test contains an illegal value, then it may not be cached
			# (since the illegal value will be repeatedly replaced with different constants), so
			# don't cache anything in that case.
			test_sigspec = self.rhs_compiler(stmt.test)

		with self._case.switch(test_sigspec, src = _src(stmt.src_loc)) as switch:
			for values, stmts in stmt.cases.items():
				case_attrs = {}
				case_src = None
				if values in stmt.case_src_locs:
					case_src = _src(stmt.case_src_locs[values])
				if isinstance(stmt.test, ast.Signal) and stmt.test.decoder:
					decoded_values = []
					for value in values:
						if '-' in value:
							decoded_values.append('<multiple>')
						else:
							decoded_values.append(stmt.test.decoder(int(value, 2)))
					case_attrs['torii.decoding'] = '|'.join(decoded_values)
				with self.case(switch, values, attrs = case_attrs, src = case_src):
					self._wrap_assign = False
					self.on_statements(stmts)
		self._wrap_assign = True

	def on_statement(self, stmt):
		try:
			super().on_statement(stmt)
		except _LegalizeValue as legalize:
			with self._case.switch(
				self.rhs_compiler(legalize.value),
				src = _src(legalize.src_loc)
			) as switch:
				shape = legalize.value.shape()
				tests = [f'{v:0{shape.width}b}' for v in legalize.branches]
				if tests:
					tests[-1] = '-' * shape.width
				for branch, test in zip(legalize.branches, tests):
					with self.case(switch, (test,)):
						self._wrap_assign = False
						branch_value = ast.Const(branch, shape)
						with self.state.expand_to(legalize.value, branch_value):
							self.on_statement(stmt)
			self._wrap_assign = True

	def on_statements(self, stmts):
		for stmt in stmts:
			self.on_statement(stmt)


def _convert_fragment(builder, fragment, name_map, hierarchy):
	if isinstance(fragment, ir.Instance):
		port_map = OrderedDict()
		for port_name, (value, dir) in fragment.named_ports.items():
			port_map[f'\\{port_name}'] = value

		if fragment.type[0] == '$':
			return fragment.type, port_map
		else:
			return f'\\{fragment.type}', port_map

	module_name  = '.'.join(name or 'anonymous' for name in hierarchy)
	module_attrs = OrderedDict()
	if len(hierarchy) == 1:
		module_attrs['top'] = 1

	with builder.module(module_name, attrs = module_attrs) as module:
		compiler_state = _ValueCompilerState(module)
		rhs_compiler   = _RHSValueCompiler(compiler_state)
		lhs_compiler   = _LHSValueCompiler(compiler_state)
		stmt_compiler  = _StatementCompiler(compiler_state, rhs_compiler, lhs_compiler)

		# Register all signals driven in the current fragment. This must be done first, as it
		# affects further codegen; e.g. whether \sig$next signals will be generated and used.
		for domain, signal in fragment.iter_drivers():
			compiler_state.add_driven(signal, sync = domain is not None)

		# Transform all signals used as ports in the current fragment eagerly and outside of
		# any hierarchy, to make sure they get sensible (non-prefixed) names.
		for signal in fragment.ports:
			compiler_state.add_port(signal, fragment.ports[signal])
			compiler_state.resolve_curr(signal)

		# Transform all clocks clocks and resets eagerly and outside of any hierarchy, to make
		# sure they get sensible (non-prefixed) names. This does not affect semantics.
		for domain, _ in fragment.iter_sync():
			cd = fragment.domains[domain]
			compiler_state.resolve_curr(cd.clk)
			if cd.rst is not None:
				compiler_state.resolve_curr(cd.rst)

		# Transform all subfragments to their respective cells. Transforming signals connected
		# to their ports into wires eagerly makes sure they get sensible (prefixed with submodule
		# name) names.
		for subfragment, sub_name in fragment.subfragments:
			if not (subfragment.ports or subfragment.statements or subfragment.subfragments):
				# If the fragment is completely empty, skip translating it, otherwise synthesis
				# tools (including Yosys and Vivado) will treat it as a black box when it is
				# loaded after conversion to Verilog.
				continue

			if sub_name is None:
				sub_name = module.anonymous()

			sub_params = OrderedDict(getattr(subfragment, 'parameters', {}))

			sub_type, sub_port_map = _convert_fragment(
				builder, subfragment, name_map,
				hierarchy = hierarchy + (sub_name,)
			)

			if sub_type == '$mem_v2' and 'MEMID' not in sub_params:
				sub_params['MEMID'] = builder._make_name(sub_name, local = False)

			sub_ports = OrderedDict()
			for port, value in sub_port_map.items():
				if not isinstance(subfragment, ir.Instance):
					for signal in value._rhs_signals():
						compiler_state.resolve_curr(signal, prefix = sub_name)
				if len(value) > 0 or sub_type == '$mem_v2':
					sub_ports[port] = rhs_compiler(value)

			module.cell(sub_type, name = sub_name, ports = sub_ports, params = sub_params,
						attrs = subfragment.attrs)

		# If we emit all of our combinatorial logic into a single RTLIL process, Verilog
		# simulators will break horribly, because Yosys write_verilog transforms RTLIL processes
		# into always @* blocks with blocking assignment, and that does not create delta cycles.
		#
		# Therefore, we translate the fragment as many times as there are independent groups
		# of signals (a group is a transitive closure of signals that appear together on LHS),
		# splitting them into many RTLIL (and thus Verilog) processes.
		lhs_grouper = xfrm.LHSGroupAnalyzer()
		lhs_grouper.on_statements(fragment.statements)

		for group, group_signals in lhs_grouper.groups().items():
			lhs_group_filter = xfrm.LHSGroupFilter(group_signals)
			group_stmts = lhs_group_filter(fragment.statements)

			with module.process(name = f'$group_{group}') as process:
				with process.case() as case:
					# For every signal in comb domain, assign \sig$next to the reset value.
					# For every signal in sync domains, assign \sig$next to the current
					# value (\sig).
					for domain, signal in fragment.iter_drivers():
						if signal not in group_signals:
							continue
						if domain is None:
							prev_value = ast.Const(signal.reset, signal.width)
						else:
							prev_value = signal
						case.assign(lhs_compiler(signal), rhs_compiler(prev_value))

					# Convert statements into decision trees.
					stmt_compiler._case = case
					stmt_compiler._has_rhs = False
					stmt_compiler._wrap_assign = False
					stmt_compiler(group_stmts)

		# For every driven signal in the sync domain, create a flop of appropriate type. Which type
		# is appropriate depends on the domain: for domains with sync reset, it is a $dff, for
		# domains with async reset it is an $adff. The latter is directly provided with the reset
		# value as a parameter to the cell, which is directly assigned during reset.
		for domain, signal in fragment.iter_sync():
			cd = fragment.domains[domain]

			wire_clk = compiler_state.resolve_curr(cd.clk)
			wire_rst = compiler_state.resolve_curr(cd.rst) if cd.rst is not None else None
			wire_curr, wire_next = compiler_state.resolve(signal)

			if not cd.async_reset:
				# For sync reset flops, the reset value comes from logic inserted by
				# `hdl.xfrm.DomainLowerer`.
				module.cell('$dff', ports = {
					'\\CLK': wire_clk,
					'\\D'  : wire_next,
					'\\Q'  : wire_curr
				}, params = {
					'CLK_POLARITY': int(cd.clk_edge == 'pos'),
					'WIDTH'       : signal.width
				})
			else:
				# For async reset flops, the reset value is provided directly to the cell.
				module.cell('$adff', ports = {
					'\\ARST': wire_rst,
					'\\CLK' : wire_clk,
					'\\D'   : wire_next,
					'\\Q'   : wire_curr
				}, params = {
					'ARST_POLARITY': ast.Const(1),
					'ARST_VALUE'   : ast.Const(signal.reset, signal.width),
					'CLK_POLARITY' : int(cd.clk_edge == 'pos'),
					'WIDTH'        : signal.width
				})

		# Any signals that are used but neither driven nor connected to an input port always
		# assume their reset values. We need to assign the reset value explicitly, since only
		# driven sync signals are handled by the logic above.
		#
		# Because this assignment is done at a late stage, a single Signal object can get assigned
		# many times, once in each module it is used. This is a deliberate decision; the possible
		# alternatives are to add ports for un-driven signals (which requires choosing one module
		# to drive it to reset value arbitrarily) or to replace them with their reset value (which
		# removes valuable source location information).
		driven = ast.SignalSet()
		for domain, signals in fragment.iter_drivers():
			driven.update(flatten(signal._lhs_signals() for signal in signals))
		driven.update(fragment.iter_ports(dir = 'i'))
		driven.update(fragment.iter_ports(dir = 'io'))
		for subfragment, sub_name in fragment.subfragments:
			driven.update(subfragment.iter_ports(dir = 'o'))
			driven.update(subfragment.iter_ports(dir = 'io'))

		for wire in compiler_state.wires:
			if wire in driven:
				continue
			wire_curr, _ = compiler_state.wires[wire]
			module.connect(wire_curr, rhs_compiler(ast.Const(wire.reset, wire.width)))

	# Collect the names we've given to our ports in RTLIL, and correlate these with the signals
	# represented by these ports. If we are a submodule, this will be necessary to create a cell
	# for us in the parent module.
	port_map = OrderedDict()
	for signal in fragment.ports:
		port_map[compiler_state.resolve_curr(signal)] = signal

	# Finally, collect the names we've given to each wire in RTLIL, and provide these to
	# the caller, to allow manipulating them in the toolchain.
	for signal in compiler_state.wires:
		wire_name = compiler_state.resolve_curr(signal)
		if wire_name.startswith('\\'):
			wire_name = wire_name[1:]
		name_map[signal] = hierarchy + (wire_name,)

	return module.name, port_map


def convert_fragment(
	fragment: ir.Fragment, name: str = 'top', *, emit_src: bool = True
) -> tuple[str, ast.SignalDict]:
	if not isinstance(fragment, ir.Fragment):
		raise ValueError(f'Expected an ir.Fragment not a {fragment!r}')
	builder = _Builder(emit_src = emit_src)
	name_map = ast.SignalDict()
	_convert_fragment(builder, fragment, name_map, hierarchy = (name,))
	return (str(builder), name_map)


def convert(
	elaboratable, name: str = 'top', platform = None, *, ports,  emit_src = True, **kwargs
) -> str:
	fragment = ir.Fragment.get(elaboratable, platform).prepare(ports = ports, **kwargs)
	il_text, _ = convert_fragment(fragment, name, emit_src = emit_src)
	return il_text



# SPDX-License-Identifier: BSD-2-Clause

import os
import re
import shutil
import subprocess
import textwrap
import traceback
from pathlib          import Path

from torii.hdl.ast    import *
from torii.hdl.ir     import *
from torii.back       import rtlil
from torii.tools      import require_tool
from torii.test       import ToriiTestCase


__all__ = (
	'ToriiTestSuiteCase',
)


class ToriiTestSuiteCase(ToriiTestCase):
	def assertRepr(self, obj, repr_str):
		if isinstance(obj, list):
			obj = Statement.cast(obj)

		def prepare_repr(repr_str):
			repr_str = re.sub(r'\s+', ' ',  repr_str)
			repr_str = re.sub(r'\( (?=\()', '(', repr_str)
			repr_str = re.sub(r'\) (?=\))', ')', repr_str)
			return repr_str.strip()
		self.assertEqual(prepare_repr(repr(obj)), prepare_repr(repr_str))

	# TODO: Once the Torii formal bits are better defined remove this and add formal to ToriiTestCase
	def assertFormal(self, spec, mode = 'bmc', depth = 1):
		stack      = traceback.extract_stack()
		file_dir   = Path(__file__).resolve().parent
		formal_dir = Path.cwd() / 'torii-formal'

		if not formal_dir.exists():
			formal_dir.mkdir(parents = True, exist_ok = True)

		for frame in reversed(stack):
			if str(file_dir) not in frame.filename:
				break
			caller = frame

		fname = Path(caller.filename).resolve().stem.replace('test_', 'spec_')
		spec_name = f'{fname}_{caller.name.replace("test_", "")}'
		spec_dir = formal_dir / spec_name

		# The sby -f switch seems not fully functional when sby is reading from stdin.
		if spec_dir.exists():
			shutil.rmtree(spec_dir)

		if mode == 'hybrid':
			# A mix of BMC and k-induction, as per personal communication with Claire Wolf.
			script = 'setattr -unset init w:* a:torii.sample_reg %d'
			mode   = 'bmc'
		else:
			script = ''

		config = textwrap.dedent(f'''\
		[options]
		mode {mode}
		depth {depth}
		wait on
		multiclock on

		[engines]
		smtbmc

		[script]
		read_rtlil top.il
		prep
		{script}

		[file top.il]
		{rtlil.convert_fragment(Fragment.get(spec, platform = 'formal').prepare())[0]}
		''')

		with subprocess.Popen([
				require_tool('sby'), '-f', '-d', spec_name
			],
			cwd                = formal_dir,
			env                = {
				**os.environ,
				'PYTHONWARNINGS': 'ignore'
			},
			universal_newlines = True,
			stdin              = subprocess.PIPE,
			stdout             = subprocess.PIPE
		) as proc:
			stdout, stderr = proc.communicate(config)
			if proc.returncode != 0:
				self.fail('Formal verification failed:\n' + stdout)



# SPDX-License-Identifier: BSD-2-Clause

from typing import Literal, Optional

from ..util import tracer
from .ast   import Signal

__all__ = (
	'ClockDomain',
	'DomainError',
)


class DomainError(Exception):
	pass


class ClockDomain:
	'''
	Synchronous domain.

	Parameters
	----------
	name : str or None
		Domain name. If ``None`` (the default) the name is inferred from the variable name this
		``ClockDomain`` is assigned to (stripping any `"cd_"` prefix).
	reset_less : bool
		If ``True``, the domain does not use a reset signal. Registers within this domain are
		still all initialized to their reset state once, e.g. through Verilog `"initial"`
		statements.
	clk_edge : str
		The edge of the clock signal on which signals are sampled. Must be one of "pos" or "neg".
	async_reset : bool
		If ``True``, the domain uses an asynchronous reset, and registers within this domain
		are initialized to their reset state when reset level changes. Otherwise, registers
		are initialized to reset state at the next clock cycle when reset is asserted.
	local : bool
		If ``True``, the domain will propagate only downwards in the design hierarchy. Otherwise,
		the domain will propagate everywhere.

	Attributes
	----------
	clk : Signal, inout
		The clock for this domain. Can be driven or used to drive other signals (preferably
		in combinatorial context).
	rst : Signal or None, inout
		Reset signal for this domain. Can be driven or used to drive.

	'''

	@staticmethod
	def _name_for(domain_name: str, signal_name: str) -> str:
		if domain_name == 'sync':
			return signal_name
		else:
			return f'{domain_name}_{signal_name}'

	def __init__(self,
		name: Optional[str] = None, *, clk_edge: Literal['pos', 'neg'] = 'pos',
		reset_less: bool = False, async_reset: bool = False, local: bool = False
	) -> None:
		if name is None:
			try:
				name = tracer.get_var_name()
			except tracer.NameNotFound:
				raise ValueError('Clock domain name must be specified explicitly')
		if name.startswith('cd_'):
			name = name[3:]
		if name == 'comb':
			raise ValueError(f'Domain \'{name}\' may not be clocked')

		if clk_edge not in ('pos', 'neg'):
			raise ValueError(f'Domain clock edge must be one of \'pos\' or \'neg\', not {clk_edge!r}')


		self.name = name

		self.clk = Signal(name = self._name_for(name, 'clk'), src_loc_at = 1)
		self.clk_edge = clk_edge

		if reset_less:
			self.rst = None
		else:
			self.rst = Signal(name = self._name_for(name, 'rst'), src_loc_at = 1)

		self.async_reset = async_reset

		self.local = local

	def rename(self, new_name: str) -> None:
		self.name = new_name
		self.clk.name = self._name_for(new_name, 'clk')
		if self.rst is not None:
			self.rst.name = self._name_for(new_name, 'rst')



# SPDX-License-Identifier: BSD-2-Clause

from typing   import Literal, Optional, Union

from ...build import Attrs, Clock, Pins, PinsN, Resource, Subsignal

__all__ = (
	'CANResource',
	'DirectUSBResource',
	'I2CResource',
	'IrDAResource',
	'JTAGResource',
	'PS2Resource',
	'SPIResource',
	'UARTResource',
	'ULPIResource',
)

def UARTResource(
	*args,
	rx: str, tx: str, rts: Optional[str] = None, cts: Optional[str] = None,
	dtr: Optional[str] = None, dsr: Optional[str] = None, dcd: Optional[str] = None,
	ri: Optional[str] = None, conn: Optional[Union[tuple[str, int], str]] = None,
	attrs: Optional[Attrs] = None, role: Optional[str]  = None
) -> Resource:

	if any(line is not None for line in (rts, cts, dtr, dsr, dcd, ri)):
		assert role in ('dce', 'dte')

	if role == 'dte':
		dce_to_dte = 'i'
		dte_to_dce = 'o'
	else:
		dce_to_dte = 'o'
		dte_to_dce = 'i'

	io = []

	io.append(Subsignal('rx', Pins(rx, dir = 'i', conn = conn, assert_width = 1)))
	io.append(Subsignal('tx', Pins(tx, dir = 'o', conn = conn, assert_width = 1)))

	if rts is not None:
		io.append(Subsignal('rts', Pins(rts, dir = dte_to_dce, conn = conn, assert_width = 1)))

	if cts is not None:
		io.append(Subsignal('cts', Pins(cts, dir = dce_to_dte, conn = conn, assert_width = 1)))

	if dtr is not None:
		io.append(Subsignal('dtr', Pins(dtr, dir = dte_to_dce, conn = conn, assert_width = 1)))

	if dsr is not None:
		io.append(Subsignal('dsr', Pins(dsr, dir = dce_to_dte, conn = conn, assert_width = 1)))

	if dcd is not None:
		io.append(Subsignal('dcd', Pins(dcd, dir = dce_to_dte, conn = conn, assert_width = 1)))

	if ri is not None:
		io.append(Subsignal('ri', Pins(ri, dir = dce_to_dte, conn = conn, assert_width = 1)))

	if attrs is not None:
		io.append(attrs)

	return Resource.family(*args, default_name = 'uart', ios = io)


def IrDAResource(
	number: int, *, rx: str, tx: str, en: Optional[str] = None, sd: Optional[str] = None,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> Resource:
	# Exactly one of en (active-high enable) or sd (shutdown, active-low enable) should
	# be specified, and it is mapped to a logic level en subsignal.
	assert (en is not None) ^ (sd is not None)

	io = []

	io.append(Subsignal('rx', Pins(rx, dir = 'i', conn = conn, assert_width = 1)))
	io.append(Subsignal('tx', Pins(tx, dir = 'o', conn = conn, assert_width = 1)))

	if en is not None:
		io.append(Subsignal('en', Pins(en, dir = 'o', conn = conn, assert_width = 1)))

	if sd is not None:
		io.append(Subsignal('en', PinsN(sd, dir = 'o', conn = conn, assert_width = 1)))

	if attrs is not None:
		io.append(attrs)

	return Resource('irda', number, *io)


def SPIResource(
	*args,
	cs_n: str, clk: str, copi: str, cipo: str, int: str = None, reset: str = None,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None,
	role: Literal['controller', 'peripheral'] = 'controller'
) -> Resource:

	if role not in ('controller', 'peripheral'):
		raise ValueError(f'role is expected to be \'controller\' or \'peripheral\', not {role!r}')

	assert copi is not None or cipo is not None # support unidirectional SPI

	io = []

	if role == 'controller':
		io.append(Subsignal('cs', PinsN(cs_n, dir = 'o', conn = conn)))
		io.append(Subsignal('clk', Pins(clk, dir = 'o', conn = conn, assert_width = 1)))
		if copi is not None:
			io.append(Subsignal('copi', Pins(copi, dir = 'o', conn = conn, assert_width = 1)))
		if cipo is not None:
			io.append(Subsignal('cipo', Pins(cipo, dir = 'i', conn = conn, assert_width = 1)))
	else:  # peripheral
		io.append(Subsignal('cs', PinsN(cs_n, dir = 'i', conn = conn, assert_width = 1)))
		io.append(Subsignal('clk', Pins(clk, dir = 'i', conn = conn, assert_width = 1)))
		if copi is not None:
			io.append(Subsignal('copi', Pins(copi, dir = 'i', conn = conn, assert_width = 1)))
		if cipo is not None:
			io.append(Subsignal('cipo', Pins(cipo, dir = 'oe', conn = conn, assert_width = 1)))

	if int is not None:
		if role == 'controller':
			io.append(Subsignal('int', Pins(int, dir = 'i', conn = conn)))
		else:
			io.append(Subsignal('int', Pins(int, dir = 'oe', conn = conn, assert_width = 1)))

	if reset is not None:
		if role == 'controller':
			io.append(Subsignal('reset', Pins(reset, dir = 'o', conn = conn)))
		else:
			io.append(Subsignal('reset', Pins(reset, dir = 'i', conn = conn, assert_width = 1)))

	if attrs is not None:
		io.append(attrs)

	return Resource.family(*args, default_name = 'spi', ios = io)


def I2CResource(
	*args,
	scl: str, sda: str, conn: Optional[Union[tuple[str, int], int]] = None,
	attrs: Optional[Attrs] = None
) -> Resource:

	io = []

	io.append(Subsignal('scl', Pins(scl, dir = 'io', conn = conn, assert_width = 1)))
	io.append(Subsignal('sda', Pins(sda, dir = 'io', conn = conn, assert_width = 1)))

	if attrs is not None:
		io.append(attrs)

	return Resource.family(*args, default_name = 'i2c', ios = io)


def DirectUSBResource(
	*args,
	d_p: str, d_n: str, pullup: Optional[str] = None, vbus_valid: Optional[str] = None,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> Resource:

	io = []

	io.append(Subsignal('d_p', Pins(d_p, dir = 'io', conn = conn, assert_width = 1)))
	io.append(Subsignal('d_n', Pins(d_n, dir = 'io', conn = conn, assert_width = 1)))

	if pullup:
		io.append(Subsignal('pullup', Pins(pullup, dir = 'o', conn = conn, assert_width = 1)))

	if vbus_valid:
		io.append(Subsignal('vbus_valid', Pins(vbus_valid, dir = 'i', conn = conn, assert_width = 1)))

	if attrs is not None:
		io.append(attrs)

	return Resource.family(*args, default_name = 'usb', ios = io)


def ULPIResource(
	*args,
	data: str, clk: str, dir: str, nxt: str, stp: str, rst: Optional[str] = None,
	clk_dir: Literal['i', 'o'] = 'i', rst_invert: bool = False, attrs: Optional[Attrs] = None,
	clk_attrs: Optional[Attrs] = None, conn: Optional[Union[tuple[str, int], int]] = None
) -> Resource:

	if clk_dir not in ('i', 'o'):
		raise ValueError(f'clk_dir should be \'i\' or \'o\' not {clk_dir!r}')

	clk_subsig = Subsignal('clk', Pins(clk, dir = clk_dir, conn = conn, assert_width = 1))
	# If the clock is an input, we must constrain it to be 60MHz.
	if clk_dir == 'i':
		clk_subsig.clock = Clock(60e6)
	if clk_attrs is not None:
		clk_subsig.attrs.update(clk_attrs)

	io = []

	io.append(Subsignal('data', Pins(data, dir = 'io', conn = conn, assert_width = 8)))
	io.append(clk_subsig)
	io.append(Subsignal('dir', Pins(dir, dir = 'i', conn = conn, assert_width = 1)))
	io.append(Subsignal('nxt', Pins(nxt, dir = 'i', conn = conn, assert_width = 1)))
	io.append(Subsignal('stp', Pins(stp, dir = 'o', conn = conn, assert_width = 1)))

	if rst is not None:
		io.append(Subsignal(
			'rst', Pins(rst, dir = 'o', invert = rst_invert, conn = conn, assert_width = 1)
		))

	if attrs is not None:
		io.append(attrs)

	return Resource.family(*args, default_name = 'usb', ios = io)


def PS2Resource(
	*args,
	clk: str, dat: str,
	conn :  Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> Resource:
	ios = []

	ios.append(Subsignal('clk', Pins(clk, dir = 'i', conn = conn, assert_width = 1))),
	ios.append(Subsignal('dat', Pins(dat, dir = 'io', conn = conn, assert_width = 1))),

	if attrs is not None:
		ios.append(attrs)

	return Resource.family(*args, default_name = 'ps2', ios = ios)

def CANResource(
	*args, rx: str, tx: str,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> Resource:
	ios = [
		Subsignal('rx', Pins(rx, dir = 'i', conn = conn)),
		Subsignal('tx', Pins(tx, dir = 'o', conn = conn)),
	]

	if attrs is not None:
		ios.append(attrs)

	return Resource.family(*args, default_name = 'can', ios = ios)

def JTAGResource(
	*args, tck: str, tms: str, tdi: str, tdo: str,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None
) -> Resource:
	ios = [
		Subsignal('tck', Pins(tck, dir = 'i', conn = conn, assert_width = 1)),
		Subsignal('tms', Pins(tms, dir = 'i', conn = conn, assert_width = 1)),
		Subsignal('tdi', Pins(tdi, dir = 'i', conn = conn, assert_width = 1)),
		Subsignal('tdo', Pins(tdo, dir = 'oe', conn = conn, assert_width = 1)),
	]
	if attrs is not None:
		ios.append(attrs)
	return Resource.family(*args, default_name = 'jtag', ios = ios)

def EthernetResource(
	*args,
	rxck: str, rxd: str, txck: str, txd: str,
	rx_dv: Optional[str] = None, rx_err: Optional[str] = None, rx_ctl: Optional[str] = None,
	tx_en: Optional[str] = None, tx_err: Optional[str] = None, tx_ctl: Optional[str] = None,
	col: Optional[str] = None, crs: Optional[str] = None,
	mdc: Optional[str] = None, mdio: Optional[str] = None,
	conn: Optional[Union[tuple[str, int], int]] = None, attrs: Optional[Attrs] = None,
	mdio_attrs: Optional[Attrs] = None
) -> Resource:

	if len(rxd.split()) not in (4, 8):
		raise ValueError(
			f'{len(rxd.split())} names are specified ({rxd}) but one of (4, 8) was expected'
		)

	if len(txd.split()) not in (4, 8):
		raise ValueError(
			f'{len(txd.split())} names are specified ({txd}) but one of (4, 8) was expected'
		)


	ios = [
		Subsignal('rx_clk', Pins(rxck, dir = 'i', conn = conn, assert_width = 1)),
		Subsignal('rx_dat', Pins(rxd, dir = 'i', conn = conn)),
		Subsignal('tx_clk', Pins(txck, dir = 'i', conn = conn, assert_width = 1)),
		Subsignal('tx_dat', Pins(txd, dir = 'o', conn = conn)),
	]

	if rx_dv is not None and rx_err is not None:
		assert rx_ctl is None
		ios.append(Subsignal('rx_dv', Pins(rx_dv, dir = 'i', conn = conn, assert_width = 1)))
		ios.append(Subsignal('rx_err', Pins(rx_err, dir = 'i', conn = conn, assert_width = 1)))
	elif rx_ctl is not None:
		ios.append(Subsignal('rx_ctl', Pins(rx_ctl, dir = 'i', conn = conn, assert_width = 1)))
	else:
		raise AssertionError('Must specify either MII RXDV + RXER pins or RGMII RXCTL')

	if tx_en is not None and tx_err is not None:
		assert tx_ctl is None
		ios.append(Subsignal('tx_en', Pins(tx_en, dir = 'o', conn = conn, assert_width = 1)))
		ios.append(Subsignal('tx_err', Pins(tx_err, dir = 'o', conn = conn, assert_width = 1)))
		if col is not None:
			ios.append(Subsignal('col', Pins(col, dir = 'i', conn = conn, assert_width = 1)))
		if crs is not None:
			ios.append(Subsignal('crs', Pins(crs, dir = 'i', conn = conn, assert_width = 1)))
	elif tx_ctl is not None:
		assert col is None and crs is None
		ios.append(Subsignal('tx_ctl', Pins(tx_ctl, dir = 'o', conn = conn, assert_width = 1)))
	else:
		raise AssertionError('Must specify either MII TXDV + TXER pins or RGMII TXCTL')

	assert (rx_dv is not None and rx_err is not None) == (tx_en is not None and tx_err is not None)
	assert (rx_ctl is not None) == (tx_ctl is not None)

	if mdc is not None and mdio is not None:
		ios.append(Subsignal('mdc', Pins(mdc, dir = 'o', conn = conn, assert_width = 1), mdio_attrs))
		ios.append(Subsignal('mdio', Pins(mdio, dir = 'io', conn = conn, assert_width = 1), mdio_attrs))
	if attrs is not None:
		ios.append(attrs)
	return Resource.family(*args, default_name = 'eth', ios = ios)



from torii.hdl.ast import *
from torii.test    import ToriiTestCase
from types         import SimpleNamespace

class TracerTestCase(ToriiTestCase):
    def test_fast(self):
        s1 = Signal()
        self.assertEqual(s1.name, "s1")
        s2 = Signal()
        self.assertEqual(s2.name, "s2")

    def test_name(self):
        class Dummy:
            s1 = Signal()
            self.assertEqual(s1.name, "s1")
            s2 = Signal()
            self.assertEqual(s2.name, "s2")

    def test_attr(self):
        ns = SimpleNamespace()
        ns.s1 = Signal()
        self.assertEqual(ns.s1.name, "s1")
        ns.s2 = Signal()
        self.assertEqual(ns.s2.name, "s2")

    def test_index(self):
        lst = [None]
        lst[0] = Signal()
        self.assertEqual(lst[0].name, "$signal")

    def test_deref_cell(self):
        s1 = Signal()
        self.assertEqual(s1.name, "s1")
        s2 = Signal()
        self.assertEqual(s2.name, "s2")

        def dummy():
            return s1, s2

    def test_deref_free(self):
        def inner():
            nonlocal s3, s4
            s3 = Signal()
            s4 = Signal()
            return s1, s2

        s1 = Signal()
        s2 = Signal()
        s3 = None
        s4 = None
        inner()
        self.assertEqual(s1.name, "s1")
        self.assertEqual(s2.name, "s2")
        self.assertEqual(s3.name, "s3")
        self.assertEqual(s4.name, "s4")

    def test_long(self):
        test = ""
        for i in range(100000):
            test += f"dummy{i} = None\n"
        test += "s1 = Signal()\n"
        test += "s2 = Signal()\n"
        ns = {"Signal": Signal}
        exec(test, ns)
        self.assertEqual(ns["s1"].name, "s1")
        self.assertEqual(ns["s2"].name, "s2")

    def test_deref_fast(self):
        def inner(s2):
            s1 = Signal()
            s2 = Signal()
            self.assertEqual(s1.name, "s1")
            self.assertEqual(s2.name, "s2")

            def dummy():
                return s1, s2

        inner(None)



# SPDX-License-Identifier: BSD-2-Clause
try:
	from importlib import metadata
	__version__ = metadata.version(__package__)
except ImportError: # pragma: no cover
	# No importlib_metadata. This shouldn't normally happen, but some people prefer not installing
	# packages via pip at all, instead using PYTHONPATH directly or copying the package files into
	# `lib/pythonX.Y/site-packages`. Although not a recommended way, we still try to support it.
	__version__ = 'unknown'


from .hdl import (
	Array, Cat, ClockDomain, ClockSignal, Const, DomainRenamer,
	Elaboratable, EnableInserter, Fragment, Instance, Memory,
	Module, Mux, Record, Repl, ResetInserter, ResetSignal, Shape,
	Signal, Value, signed, unsigned
)

__all__ = (
	'Array',
	'Cat',
	'ClockDomain',
	'ClockSignal',
	'Const',
	'DomainRenamer',
	'Elaboratable',
	'EnableInserter',
	'Fragment',
	'Instance',
	'Memory',
	'Module',
	'Mux',
	'Record',
	'Repl',
	'ResetInserter',
	'ResetSignal',
	'Shape',
	'Signal',
	'Value',
	'signed',
	'unsigned',
)



# SPDX-License-Identifier: BSD-2-Clause

from typing        import Literal, Optional, Union

from ...           import Cat, Const, Elaboratable, Module, Record, Signal
from ...util.units import bits_for
from ..cdc         import FFSynchronizer
from ..io          import Pin

__all__ = (
	'AsyncSerial',
	'AsyncSerialRX',
	'AsyncSerialTX',
)


def _check_divisor(divisor: int, bound: int):
	if divisor < bound:
		raise ValueError(f'Invalid divisor {divisor!r}; must be greater than or equal to {bound}')


def _check_parity(parity: Literal['none', 'mark', 'space', 'even', 'odd']):
	choices = ( 'none', 'mark', 'space', 'even', 'odd' )
	if parity not in choices:
		raise ValueError(f'Invalid parity {parity!r}; must be one of {", ".join(choices)}')


def _compute_parity_bit(
	data: Record , parity: Literal['none', 'mark', 'space', 'even', 'odd']
) -> Union[Const, Record, bool]:
	if parity == 'none':
		return Const(0, 0)
	if parity == 'mark':
		return Const(1, 1)
	if parity == 'space':
		return Const(0, 1)
	if parity == 'even':
		return data.xor()
	if parity == 'odd':
		return ~data.xor()
	assert False


def _wire_layout(
	data_bits: int, parity: Literal['none', 'mark', 'space', 'even', 'odd'] = 'none'
) -> list[tuple[str, int]]:
	return [
		('start',  1),
		('data',   data_bits),
		('parity', 0 if parity == 'none' else 1),
		('stop',   1),
	]


class AsyncSerialRX(Elaboratable):
	'''
	Asynchronous serial receiver.

	Parameters
	----------
	divisor : int
		Clock divisor reset value. Should be set to ``int(clk_frequency // baudrate)``.
	divisor_bits : int
		Optional. Clock divisor width. If omitted, ``bits_for(divisor)`` is used instead.
	data_bits : int
		Data width.
	parity : ``'none'``, ``'mark'``, ``'space'``, ``'even'``, ``'odd'``
		Parity mode.
	pins : :class:`torii.lib.io.Pin`
		Optional. UART pins. See :class:`torii_boards.resources.UARTResource` for layout.

	Attributes
	----------
	divisor : Signal, in
		Clock divisor.
	data : Signal, out
		Read data. Valid only when ``rdy`` is asserted.
	err.overflow : Signal, out
		Error flag. A new frame has been received, but the previous one was not acknowledged.
	err.frame : Signal, out
		Error flag. The received bits do not fit in a frame.
	err.parity : Signal, out
		Error flag. The parity check has failed.
	rdy : Signal, out
		Read strobe.
	ack : Signal, in
		Read acknowledge. Must be held asserted while data can be read out of the receiver.
	i : Signal, in
		Serial input. If ``pins`` has been specified, ``pins.rx.i`` drives it.

	'''

	def __init__(
		self, *, divisor: int, divisor_bits: Optional[int] = None, data_bits: int = 8,
		parity: Literal['none', 'mark', 'space', 'even', 'odd'] = 'none',
		pins: Optional[Pin] = None
	) -> None:
		_check_parity(parity)
		self._parity = parity

		# The clock divisor must be at least 5 to keep the FSM synchronized with the serial input
		# during a DONE->IDLE->BUSY transition.
		_check_divisor(divisor, 5)
		self.divisor = Signal(divisor_bits or bits_for(divisor), reset = divisor)

		self.data = Signal(data_bits)
		self.err  = Record([
			('overflow', 1),
			('frame',    1),
			('parity',   1),
		])
		self.rdy  = Signal()
		self.ack  = Signal()

		self.i    = Signal(reset = 1)

		self._pins = pins

	def elaborate(self, platform) -> Module:
		m = Module()

		timer = Signal.like(self.divisor)
		shreg = Record(_wire_layout(len(self.data), self._parity))
		bitno = Signal(range(len(shreg)))

		if self._pins is not None:
			m.submodules += FFSynchronizer(self._pins.rx.i, self.i, reset = 1)

		with m.FSM() as fsm:
			with m.State('IDLE'):
				with m.If(~self.i):
					m.d.sync += [
						bitno.eq(len(shreg) - 1),
						timer.eq(self.divisor >> 1),
					]
					m.next = 'BUSY'

			with m.State('BUSY'):
				with m.If(timer != 0):
					m.d.sync += timer.eq(timer - 1)
				with m.Else():
					m.d.sync += [
						shreg.eq(Cat(shreg[1:], self.i)),
						bitno.eq(bitno - 1),
						timer.eq(self.divisor - 1),
					]
					with m.If(bitno == 0):
						m.next = 'DONE'

			with m.State('DONE'):
				with m.If(self.ack):
					m.d.sync += [
						self.data.eq(shreg.data),
						self.err.frame .eq(~((shreg.start == 0) & (shreg.stop == 1))),
						self.err.parity.eq(~(
							shreg.parity == _compute_parity_bit(shreg.data, self._parity)
						)),
					]
				m.d.sync += self.err.overflow.eq(~self.ack)
				m.next = 'IDLE'

		with m.If(self.ack):
			m.d.sync += self.rdy.eq(fsm.ongoing('DONE'))

		return m


class AsyncSerialTX(Elaboratable):
	'''
	Asynchronous serial transmitter.

	Parameters
	----------
	divisor : int
		Clock divisor reset value. Should be set to ``int(clk_frequency // baudrate)``.
	divisor_bits : int
		Optional. Clock divisor width. If omitted, ``bits_for(divisor)`` is used instead.
	data_bits : int
		Data width.
	parity : ``'none'``, ``'mark'``, ``'space'``, ``'even'``, ``'odd'``
		Parity mode.
	pins : :class:`torii.lib.io.Pin`
		Optional. UART pins. See :class:`torii_boards.resources.UARTResource` for layout.

	Attributes
	----------
	divisor : Signal, in
		Clock divisor.
	data : Signal, in
		Write data. Valid only when ``ack`` is asserted.
	rdy : Signal, out
		Write ready. Asserted when the transmitter is ready to transmit data.
	ack : Signal, in
		Write strobe. Data gets transmitted when both ``rdy`` and ``ack`` are asserted.
	o : Signal, out
		Serial output. If ``pins`` has been specified, it drives ``pins.tx.o``.

	'''

	def __init__(
		self, *, divisor: int, divisor_bits: Optional[int] = None, data_bits: int = 8,
		parity  : Literal['none', 'mark', 'space', 'even', 'odd'] = 'none',
		pins: Optional[Pin] = None
	) -> None:
		_check_parity(parity)
		self._parity = parity

		_check_divisor(divisor, 1)
		self.divisor = Signal(divisor_bits or bits_for(divisor), reset = divisor)

		self.data = Signal(data_bits)
		self.rdy  = Signal()
		self.ack  = Signal()

		self.o    = Signal(reset = 1)

		self._pins = pins

	def elaborate(self, platform) -> Module:
		m = Module()

		timer = Signal.like(self.divisor)
		shreg = Record(_wire_layout(len(self.data), self._parity))
		bitno = Signal(range(len(shreg)))

		if self._pins is not None:
			m.d.comb += self._pins.tx.o.eq(self.o)

		with m.FSM():
			with m.State('IDLE'):
				m.d.comb += self.rdy.eq(1)
				with m.If(self.ack):
					m.d.sync += [
						shreg.start .eq(0),
						shreg.data  .eq(self.data),
						shreg.parity.eq(_compute_parity_bit(self.data, self._parity)),
						shreg.stop  .eq(1),
						bitno.eq(len(shreg) - 1),
						timer.eq(self.divisor - 1),
					]
					m.next = 'BUSY'

			with m.State('BUSY'):
				with m.If(timer != 0):
					m.d.sync += timer.eq(timer - 1)
				with m.Else():
					m.d.sync += [
						Cat(self.o, shreg).eq(shreg),
						bitno.eq(bitno - 1),
						timer.eq(self.divisor - 1),
					]
					with m.If(bitno == 0):
						m.next = 'IDLE'

		return m


class AsyncSerial(Elaboratable):
	'''
	Asynchronous serial transceiver.

	Parameters
	----------
	divisor : int
		Clock divisor reset value. Should be set to ``int(clk_frequency // baudrate)``.
	divisor_bits : int
		Optional. Clock divisor width. If omitted, ``bits_for(divisor)`` is used instead.
	data_bits : int
		Data width.
	parity : ``'none'``, ``'mark'``, ``'space'``, ``'even'``, ``'odd'``
		Parity mode.
	pins : :class:`torii.lib.io.Pin`
		Optional. UART pins. See :class:`torii_boards.resources.UARTResource` for layout.

	Attributes
	----------
	divisor : Signal, in
		Clock divisor.
	rx : :class:`AsyncSerialRX`
		See :class:`AsyncSerialRX`.
	tx : :class:`AsyncSerialTX`
		See :class:`AsyncSerialTX`.

	'''

	def __init__(
		self, *, divisor: int, divisor_bits: Optional[int] = None, **kwargs
	) -> None:
		self.divisor = Signal(divisor_bits or bits_for(divisor), reset = divisor)

		self.rx = AsyncSerialRX(divisor = divisor, divisor_bits = divisor_bits, **kwargs)
		self.tx = AsyncSerialTX(divisor = divisor, divisor_bits = divisor_bits, **kwargs)

	def elaborate(self, platform) -> Module:
		m = Module()
		m.submodules.rx = self.rx
		m.submodules.tx = self.tx
		m.d.comb += [
			self.rx.divisor.eq(self.divisor),
			self.tx.divisor.eq(self.divisor),
		]
		return m



# SPDX-License-Identifier: BSD-2-Clause

from typing import IO, Iterable, Optional, Union

from ..hdl  import Signal

__all__ = (
	'BaseEngine',
	'BaseProcess',
	'BaseSignalState',
	'BaseSimulation',
)

class BaseProcess:
	__slots__ = ()

	def __init__(self) -> None:
		self.reset()

	def reset(self) -> None:
		self.runnable = False
		self.passive  = True

	def run(self) -> None:
		raise NotImplementedError


class BaseSignalState:
	__slots__ = ()

	signal = NotImplemented

	curr = NotImplemented
	next = NotImplemented

	def set(self, value) -> None:
		raise NotImplementedError


class BaseSimulation:
	def reset(self) -> None:
		raise NotImplementedError

	def get_signal(self, signal) -> None:
		raise NotImplementedError

	slots = NotImplemented

	def add_trigger(self, process, signal, *, trigger = None):
		raise NotImplementedError

	def remove_trigger(self, process, signal):
		raise NotImplementedError

	def wait_interval(self, process, interval):
		raise NotImplementedError


class BaseEngine:
	def add_coroutine_process(self, process, *, default_cmd):
		raise NotImplementedError

	def add_clock_process(self, clock, *, phase, period):
		raise NotImplementedError

	def reset(self) -> None:
		raise NotImplementedError

	@property
	def now(self) -> int:
		raise NotImplementedError

	def advance(self) -> bool:
		raise NotImplementedError

	def write_vcd(
		self, *, vcd_file: Optional[Union[IO, str]], gtkw_file: Optional[Union[IO, str]] = None,
		traces: Iterable[Signal]
	) -> None:
		raise NotImplementedError
